import logging
import logging.config
from pathlib import Path

import eascheduler
import pydantic

from HABApp import __version__
from HABApp.config.config import CONFIG
from HABApp.config.logging import HABAppQueueHandler, load_logging_file
from HABApp.core import shutdown
from HABApp.core.internals.proxy.proxies import uses_file_manager

from .debug import setup_debug
from .errors import AbsolutePathExpected, InvalidConfigError
from .logging import create_default_logfile, get_logging_dict
from .logging.buffered_logger import BufferedLogger


log = logging.getLogger('HABApp.Config')


file_manager = uses_file_manager()


def setup_habapp_configuration(config_folder: Path) -> None:

    CONFIG.set_file_path(config_folder / 'config.yml')
    preprocess = CONFIG.load_preprocess
    preprocess.set_log_func(log.warning)
    # old sse event handler config, remove 2026
    preprocess.delete_entry(('openhab', 'connection', 'buffer'))
    preprocess.delete_entry(('openhab', 'connection', 'topic filter'))
    # change name for param folder, remove 2026
    preprocess.move_entry(('directories', 'param'), ('directories', 'params'))

    logging_cfg_path = config_folder / 'logging.yml'
    create_default_logfile(logging_cfg_path)

    loaded_logging = False

    # Try load the logging config
    try:
        load_logging_cfg(logging_cfg_path)
        loaded_logging = True
    except (AbsolutePathExpected, InvalidConfigError):
        pass

    load_habapp_cfg(do_print=not loaded_logging)

    if not loaded_logging:
        load_logging_cfg(logging_cfg_path)

    shutdown.register(stop_queue_handlers, msg='Stop logging queue handlers', last=True)

    setup_debug()

    watcher = file_manager.get_file_watcher()
    watcher.watch_file('config.log_file', config_file_changed, config_folder / 'logging.yml', habapp_internal=True)
    watcher.watch_file('config.cfg_file', config_file_changed, config_folder / 'config.yml', habapp_internal=True)

    CONFIG.habapp.logging.subscribe_for_changes(set_flush_delay)


def set_flush_delay() -> None:
    HABAppQueueHandler.FLUSH_DELAY = CONFIG.habapp.logging.flush_every


async def config_file_changed(path: str) -> None:
    file = Path(path)
    if file.name == 'config.yml':
        load_habapp_cfg()
    if file.name == 'logging.yml':
        load_logging_cfg(file)


def load_habapp_cfg(do_print=False) -> None:
    def error(text: str) -> None:
        if do_print:
            print(text)
        else:
            log.error(text)

    try:
        CONFIG.load_config_file()
    except pydantic.ValidationError as e:
        for line in str(e).splitlines():
            error(line)
        raise InvalidConfigError from None

    # check if folders exist and print warnings, maybe because of missing permissions
    if not CONFIG.directories.rules.is_dir():
        log.warning(f'Folder for rules files does not exist: {CONFIG.directories.rules}')

    CONFIG.directories.create_folders()

    location = CONFIG.location
    eascheduler.set_location(location.latitude, location.longitude, location.elevation)

    if not location.country:
        log.warning('No country is set in the config file. Holidays will not be available.')
    else:
        try:
            eascheduler.setup_holidays(location.country, location.subdivision if location.subdivision else None)
        except Exception as e:
            for line in str(e).splitlines():
                error(line)

    log.debug('Loaded HABApp config')


QUEUE_HANDLER: list['HABAppQueueHandler'] = []


def stop_queue_handlers() -> None:
    for qh in QUEUE_HANDLER:
        qh.signal_stop()
    while QUEUE_HANDLER:
        qh = QUEUE_HANDLER.pop()
        qh.stop()


def load_logging_cfg(path: Path) -> None:
    # If the logging file gets accidentally deleted we do nothing
    if (logging_yaml := load_logging_file(path)) is None:
        return None

    # stop buffered handlers
    stop_queue_handlers()

    buf_log = BufferedLogger()
    cfg, q_handlers = get_logging_dict(logging_yaml, buf_log)

    # load prepared logging
    try:
        logging.config.dictConfig(cfg)
    except Exception as e:
        print(f'Error loading logging config: {e}')
        log.error(f'Error loading logging config: {e}')
        raise InvalidConfigError from None

    # start buffered handlers
    for qh in q_handlers:
        QUEUE_HANDLER.append(qh)
        qh.start()

    logging.getLogger('HABApp').info(f'HABApp Version {__version__}')

    # write buffered messages
    buf_log.flush(log)
    return None



from HABApp.core.connections._definitions import CONNECTION_HANDLER_NAME, ConnectionStatus


# isort: split

from HABApp.core.connections.base_connection import BaseConnection
from HABApp.core.connections.base_plugin import BaseConnectionPlugin
from HABApp.core.connections.plugin_callback import PluginCallbackHandler


# isort: split

from HABApp.core.connections.manager import connection_manager as Connections


# isort: split

from HABApp.core.connections.plugins import AutoReconnectPlugin, ConnectionStateToEventBusPlugin



from .exceptions import HINT_EXCEPTION, format_exception
from .helper import get_obj_name
from .instant_view import InstantView
from .pending_future import PendingFuture
from .priority_list import PriorityList
from .single_task import SingleTask
from .timeout import Timeout, TimeoutNotRunningError
from .value_change import ValueChange



from . import habapp_events
from .events import (
    ItemNoChangeEvent,
    ItemNoUpdateEvent,
    ValueChangeEvent,
    ValueCommandEvent,
    ValueUpdateEvent,
)
from .filter import (
    AndFilterGroup,
    EventFilter,
    NoEventFilter,
    OrFilterGroup,
    ValueChangeEventFilter,
    ValueCommandEventFilter,
    ValueUpdateEventFilter,
)



from typing import Any, Final

from typing_extensions import Self, override

from HABApp.openhab.definitions.websockets import ItemCommandSendEvent, ItemStateSendEvent, item_value_types
from HABApp.openhab.definitions.websockets.base import BaseOutEvent
from HABApp.openhab.definitions.websockets.item_value_types import ItemValueBase


class OutgoingEventBase:
    __slots__ = ('_event', '_name', '_types')

    def __init__(self, name: str, *types: str | type[ItemValueBase],
                 event: type[ItemCommandSendEvent | ItemStateSendEvent]) -> None:
        self._event: Final = event
        self._name: Final = name
        self._types: Final = tuple(self._check_type(t, check_existing=False) for t in types)

        if not issubclass(event, BaseOutEvent):
            msg = f'Expected {BaseOutEvent.__name__}, got {event}'
            raise TypeError(msg)

        if len(set(self._types)) != len(self._types):
            msg = 'Duplicate entries in types'
            raise ValueError(msg)

    def _check_type(self, other: Any, *, check_existing: bool = True) -> type[ItemValueBase]:
        if isinstance(other, str):
            suffix = 'TypeModel'
            available = {
                name.removesuffix(suffix): getattr(item_value_types, name)
                for name in dir(item_value_types) if name.endswith(suffix)
            }
            if (other := available.get(other)) is None:
                msg = f'Unknown type: {other}! Available: {", ".join(sorted(available))}'
                raise ValueError(msg)

        if not issubclass(other, ItemValueBase):
            msg = f'Expected ItemValueBase, got {other}'
            raise TypeError(msg)

        if check_existing and other in self._types:
            msg = f'{other.__name__} already in {self._name:s}'
            raise ValueError(msg)
        return other

    def __or__(self, other: type[ItemValueBase]) -> Self:
        to_add = self._check_type(other)
        return self.__class__(self._name, *self._types, to_add)

    def supported_event_names(self) -> tuple[str, ...]:
        return tuple(t.__name__ for t in self._types)

    def create_event(self, name: str, value: Any) -> BaseOutEvent:
        for t in self._types:
            if (r := t.from_value(value)) is not None:
                return self._event.create(name, r)

        if isinstance(value, ItemValueBase):
            return self._event.create(name, value)

        msg = f"Invalid value: '{value}' ({type(value)}) for {self._name:s}"
        raise ValueError(msg)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}{self._name:s}({", ".join(self.supported_event_names())})'


class OutgoingStateEvent(OutgoingEventBase):

    @override
    def __init__(self, name: str, *types: str | type[ItemValueBase]) -> None:
        super().__init__(name, *types, event=ItemStateSendEvent)


class OutgoingCommandEvent(OutgoingEventBase):
    @override
    def __init__(self, name: str, *types: str | type[ItemValueBase]) -> None:
        super().__init__(name, *types, event=ItemCommandSendEvent)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from HABApp.core.errors import InvalidItemValueError, ItemValueIsNoneError
from HABApp.openhab.definitions.websockets.item_value_types import OnOffTypeModel, RefreshTypeModel, UnDefTypeModel
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent
from HABApp.openhab.items.commands import OnOffCommand


if TYPE_CHECKING:
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/SwitchItem.java
class SwitchItem(OpenhabItem, OnOffCommand):
    """SwitchItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar str value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('SwitchItem', OnOffTypeModel, UnDefTypeModel)
    _command_to_oh: Final = OutgoingCommandEvent('SwitchItem', OnOffTypeModel, RefreshTypeModel)
    _state_from_oh_str: Final = staticmethod(OnOffTypeModel.get_value_from_state)

    def set_value(self, new_value: str | None) -> bool:
        if new_value not in ('ON', 'OFF', None):
            raise InvalidItemValueError.from_item(self, new_value)

        return super().set_value(new_value)

    def is_on(self) -> bool:
        """Test value against on-value"""
        return self.value == 'ON'

    def is_off(self) -> bool:
        """Test value against off-value"""
        return self.value == 'OFF'

    def toggle(self):
        """Toggle the switch. Turns the switch on when off or off when currently on."""
        if self.value == 'ON':
            self.off()
        elif self.value == 'OFF':
            self.on()
        elif self.value is None:
            raise ItemValueIsNoneError.from_item(self)
        else:
            raise InvalidItemValueError.from_item(self, self.value)

    def __str__(self) -> str:
        return str(self.value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SwitchItem):
            return self.value == other.value
        if isinstance(other, str):
            return self.value == other
        if isinstance(other, int):
            if other and self.is_on():
                return True
            if not other and self.is_off():
                return True
            return False

        return NotImplemented

    def __bool__(self) -> bool:
        if self.value is None:
            raise ItemValueIsNoneError.from_item(self)
        return self.value == 'ON'



from __future__ import annotations

import logging
from asyncio import sleep

from HABApp.core import shutdown
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.internals import uses_item_registry
from HABApp.core.lib import ValueChange
from HABApp.core.lib.timeout import Timeout
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.items import OpenhabItem


log = logging.getLogger('HABApp.openhab.startup')

item_registry = uses_item_registry()


def count_none_items() -> int:
    found = 0
    for item in item_registry.get_items():
        if isinstance(item, OpenhabItem) and item.value is None:
            found += 1
    return found


class WaitForPersistenceRestore(BaseConnectionPlugin[OpenhabConnection]):

    async def on_connected(self, context: OpenhabContext):
        if not context.waited_for_openhab:
            log.debug('Openhab has already been running -> complete')
            return None

        none_items: ValueChange[int] = ValueChange()

        # if we find None items check if they are still getting initialized (e.g. from persistence)
        if none_items.set_value(count_none_items()).value:
            log.debug('Some items are still None - waiting for initialisation')

            timeout = Timeout(4 * 60)
            while not shutdown.is_requested() and none_items.changed:
                await sleep(3)

                # timeout so we start eventually
                if timeout.is_expired():
                    log.debug('Timeout while waiting for initialisation')
                    break

                none_items.set_value(count_none_items())

        log.debug('complete')



from collections.abc import Mapping
from typing import Final, NoReturn, TypeAlias

from HABApp.openhab.errors import MapTransformationError


MapKeyType: TypeAlias = int | str
MapValueType: TypeAlias = int | str


class MapTransformation(dict[MapKeyType, MapValueType]):
    def __init__(self, obj: Mapping[MapKeyType, MapValueType], *, name: str) -> None:
        super().__init__(obj)
        self._name: Final = name

    def __repr__(self, additional: str = '') -> str:
        return f'<{self.__class__.__name__} name={self._name} items={super().__repr__()}{additional:s}>'


class MapTransformationWithDefault(MapTransformation):
    def __init__(self, obj: Mapping[MapKeyType, MapValueType], *, default: MapValueType, name: str) -> None:
        super().__init__(obj, name=name)
        self._default: Final = default

    def __missing__(self, key: MapKeyType) -> int | str:
        return self._default

    def __repr__(self, additional: str = '') -> str:
        return super().__repr__(f', default={self._default}{additional}')

    def get(self, key: MapKeyType, default: MapValueType | None = None) -> NoReturn:  # noqa: ARG002
        msg = f'Mapping is already defined with a default: "{self._default}"'
        raise MapTransformationError(msg)



from __future__ import annotations

import logging
from asyncio import Queue, sleep
from base64 import b64encode
from inspect import isclass
from typing import Annotated, Final, get_args, get_origin

from aiohttp import BasicAuth, ClientError, ClientWebSocketResponse, WSMsgType
from pydantic import ValidationError

import HABApp
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.const.const import PYTHON_311
from HABApp.core.const.log import TOPIC_EVENTS
from HABApp.core.internals import uses_item_registry
from HABApp.core.lib import SingleTask
from HABApp.core.logger import HABAppError, HABAppWarning
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.definitions.websockets import (
    OPENHAB_EVENT_TYPE,
    OPENHAB_EVENT_TYPE_ADAPTER,
    WebsocketHeartbeatEvent,
    WebsocketSendTypeFilter,
    WebsocketTopicEnum,
)
from HABApp.openhab.definitions.websockets.base import BaseModel, BaseOutEvent
from HABApp.openhab.process_events import on_openhab_event


if PYTHON_311:
    from asyncio import TaskGroup
else:
    from taskgroup import TaskGroup


class WebSocketClosedError(ClientError):
    pass


Items = uses_item_registry()


class WebsocketPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self, name: str | None = None) -> None:
        super().__init__(name)
        self.task: Final = SingleTask(self.websockets_task, name='WebsocketsEventsTask')

        self._websocket: ClientWebSocketResponse | None = None
        self.queue: Queue[BaseOutEvent] | None = None

        self._sent_events: Final[dict[str, BaseOutEvent]] = {}

    async def on_connected(self, context: OpenhabContext) -> None:
        self.queue: Queue[BaseOutEvent] = context.out_queue
        self.task.start()

    async def on_disconnected(self) -> None:
        await self.task.cancel_wait()
        self._sent_events.clear()

    @staticmethod
    def _build_token(auth: BasicAuth) -> str:
        login = auth.login
        password = auth.password

        # we use token as auth
        for v in (login, password):
            if v.startswith('oh.'):
                return v

        # basic auth
        return b64encode(f'{login}:{password}'.encode()).decode()

    async def _websocket_sender(self, ws: ClientWebSocketResponse, queue: Queue[BaseOutEvent]) -> None:
        while True:
            try:
                while True:
                    event = await queue.get()
                    text = event.model_dump_json(by_alias=True, exclude_none=True)
                    if (key := event.event_id) is not None:
                        self._sent_events[key] = event
                    await ws.send_str(text)
                    queue.task_done()
            except Exception as e:  # noqa: PERF203
                self.plugin_connection.process_exception(e, 'Outgoing queue worker')

    async def _websocket_ping(self, ping_interval: float) -> None:
        log = self.plugin_connection.log
        log.debug('Websocket ping task started')

        ping_text: Final = WebsocketHeartbeatEvent(
            type='WebSocketEvent',
            topic='openhab/websocket/heartbeat',
            payload='PING',
        ).model_dump_json(by_alias=True, exclude_none=True)

        stop_reason: str = ''
        msgs_last: frozenset[str] = frozenset()

        try:
            while True:
                await sleep(ping_interval)

                if (ws := self._websocket) is None:
                    stop_reason = ' (is None)'
                    return None
                await ws.send_str(ping_text)

                # Remove stale events where we didn't get an answer
                in_flight = frozenset(self._sent_events)
                if to_remove := msgs_last & in_flight:
                    wl = HABAppWarning(log).add(f'Removing events in flight since {ping_interval}s:')
                    for msg in to_remove:
                        obj = self._sent_events.pop(msg)
                        wl.add(f' - {obj.model_dump_json(by_alias=True, exclude_none=True)}')
                    wl.dump()
                msgs_last = in_flight

        except Exception as e:
            stop_reason = f' {e} ({type(e)})'
            raise
        finally:
            log.debug(f'Websocket ping task stopped{stop_reason:s}')

    async def websockets_task(self) -> None:
        try:
            session = self.plugin_connection.context.session
            token = self._build_token(session.auth)

            ws_cfg = HABApp.CONFIG.openhab.connection.websocket
            max_msg_size = int(ws_cfg.max_msg_size)
            ping_interval = ws_cfg.ping_interval

            async with session.ws_connect(
                    f'/ws?accessToken={token:s}', autoping=False, max_msg_size=max_msg_size) as ws:

                self._websocket = ws
                try:
                    async with TaskGroup() as tg:
                        tg.create_task(self._websocket_sender(ws, self.queue))
                        tg.create_task(self._websocket_ping(ping_interval))
                        tg.create_task(self._websocket_events())
                finally:
                    self._websocket = None

        # TODO: replace this with exception group as soon as 3.14 is available
        except Exception as e:
            self.plugin_connection.process_exception(e, self.websockets_task)

    @staticmethod
    def _get_type_names_from_adapter() -> list[str]:
        names = set()

        objs = [get_args(OPENHAB_EVENT_TYPE)[0]]

        while objs:
            obj = objs.pop(0)

            if isclass(obj) and issubclass(obj, BaseModel):
                literal = obj.model_fields['type'].annotation
                literal_value = get_args(literal)
                if len(literal_value) != 1:
                    msg = f'Expected exactly one literal value for {literal!r}'
                    raise ValueError(msg)
                names.add(literal_value[0])
                continue

            if get_origin(obj) is Annotated:
                objs.append(get_args(obj)[0])
                continue

            new = get_args(obj)
            if not new:
                msg = f'Expected args for {obj!r}'
                raise ValueError(msg)
            objs.extend(new)

        return sorted(names)

    async def _setup_websocket_filter(self, ws: ClientWebSocketResponse, log: logging.Logger) -> None:
        # setup event type filter
        filter_cfg = HABApp.CONFIG.openhab.connection.websocket.event_filter
        supported_event_names = set(self._get_type_names_from_adapter())

        names: set[str] = set()
        if filter_cfg.event_type.is_auto():
            names.update(supported_event_names)

            # We ignore the ItemStateEvent because we will process the ItemStateUpdatedEvent
            # which has the correct value type for the item
            names.discard('ItemStateEvent')

            # https://github.com/spacemanspiff2007/HABApp/issues/437
            # https://github.com/spacemanspiff2007/HABApp/issues/449
            # openHAB will automatically restore the future states of the item
            # which means we can safely ignore these events because we will see the ItemStateUpdatedEvent
            names.discard('ItemTimeSeriesUpdatedEvent')
            names.discard('ItemTimeSeriesEvent')

        elif filter_cfg.event_type.is_config():
            names.update(filter_cfg.types_allowed)
            if invalid := (names - supported_event_names):
                log.warning(
                    f'Invalid event type name{"s" if len(invalid) != 1 else ""} '
                    f'in filter config: {", ".join(sorted(invalid))}')
                names -= invalid

        if names:
            msg = WebsocketSendTypeFilter(payload=sorted(names))
            log.debug(f'Send: {msg.model_dump_json(by_alias=True)}')
            await ws.send_str(msg.model_dump_json(by_alias=True))

        return None

    async def _websocket_events(self) -> None:
        if (ws := self._websocket) is None:
            return None

        # cache so we don't have to look up every event
        _on_openhab_event = on_openhab_event
        log = self.plugin_connection.log
        log_events = logging.getLogger(f'{TOPIC_EVENTS}.openhab')
        debug_lvl = logging.DEBUG

        # Setup event filter
        await self._setup_websocket_filter(ws, log)

        # Websocket constants
        ws_type_text = WSMsgType.TEXT
        ws_type_close = WSMsgType.CLOSED
        topic_heartbeat = WebsocketTopicEnum.HEARTBEAT
        topic_error = WebsocketTopicEnum.REQUEST_FAILED
        topic_success = WebsocketTopicEnum.REQUEST_SUCCESS

        while True:
            msg = await ws.receive()
            msg_type, data, extra = msg

            if msg_type == ws_type_close:
                log.debug(f'Websocket closed: {ws.close_code} {extra}')
                break

            if msg_type != ws_type_text:
                log.warning(f'Message with unexpected type received: {msg}')
                continue

            if log_events.isEnabledFor(debug_lvl):
                log_events._log(debug_lvl, data, ())

            try:
                oh_event = OPENHAB_EVENT_TYPE_ADAPTER.validate_json(data)
            except ValidationError as e:
                HABAppError(log).add(f'Input: {data:s}').add_exception(e).dump()
                continue

            # Websocket events are not processed by the event bus
            if oh_event.type == 'WebSocketEvent':
                topic = oh_event.topic

                # confirmation that the event was processed
                if topic == topic_success:
                    self._sent_events.pop(oh_event.event_id, None)
                    continue

                # Error processing the sent event
                if topic == topic_error:
                    err_log = HABAppError(log)
                    err_log.add('Request failed!')
                    if (send_obj := self._sent_events.get(oh_event.event_id)) is not None:
                        err_log.add(f'Sent    : {send_obj.model_dump_json(by_alias=True, exclude_none=True):s}')
                    err_log.add(f'Received: {data}').add(f'{oh_event}').dump()
                    self._sent_events.pop(oh_event.event_id, None)
                    continue

                if topic == topic_heartbeat:
                    continue

                log.debug(f'Receive: {data}')
                log.debug(str(oh_event))
                continue

            try:
                event = oh_event.to_event()
            except ValueError as e:
                HABAppError(log).add(f'Input: {data:s}').add(f'{e} ({type(e)}').dump()
                continue

            _on_openhab_event(event)

        # We need to raise an error otherwise the task group will not exit
        raise WebSocketClosedError()



import re

from pydantic import BaseModel, ConfigDict, Field

from HABApp.core.const import yml


class FileProperties(BaseModel):
    depends_on: list[str] = Field(alias='depends on', default_factory=list)
    reloads_on: list[str] = Field(alias='reloads on', default_factory=list)

    model_config = ConfigDict(extra='forbid', populate_by_name=True)


RE_START = re.compile(r'^(\s*#\s*)HABApp\s*:', re.IGNORECASE)


def get_file_properties(_str: str) -> FileProperties:

    cfg = []
    cut = 0

    # extract the property string
    for line in _str.splitlines():
        line_strip = line.strip()
        if cut and not line_strip:
            break

        if not line_strip:
            continue

        # break on first non-empty line that is not a comment
        if line_strip and not line_strip.startswith('#'):
            break

        if not cut:
            # find out how much from the start we have to cut
            if m := RE_START.search(line):
                cut = m.end(1)
                cfg.append(line[cut:].lower())
        else:
            # If we would cut away characters it's not the yaml definition any more
            # Here it's cut + 1 because it must be indented
            if line[:cut + 1].strip() not in ('', '#'):
                break
            cfg.append(line[cut:])

    data = yml.load('\n'.join(cfg))
    if data is None:
        data = {}
    return FileProperties.model_validate(data.get('habapp', {}))



from HABApp.core.const.hints import HasNameAttr as _HasNameAttr
from HABApp.core.errors import InvalidItemValueError
from HABApp.openhab.connection.plugins import send_websocket_event
from HABApp.openhab.definitions.websockets import ItemCommandSendEvent
from HABApp.openhab.definitions.websockets.item_value_types import OnOffTypeModel, PercentTypeModel, UpDownTypeModel


class OnOffCommand:

    def is_on(self: _HasNameAttr) -> bool:
        """Test value against on-value"""
        raise NotImplementedError()

    def is_off(self: _HasNameAttr) -> bool:
        """Test value against off-value"""
        raise NotImplementedError()

    def on(self: _HasNameAttr) -> None:
        """Command item on"""
        send_websocket_event(
            ItemCommandSendEvent.create(self.name, OnOffTypeModel(type='OnOff', value='ON'))
        )

    def off(self: _HasNameAttr) -> None:
        """Command item off"""
        send_websocket_event(
            ItemCommandSendEvent.create(self.name, OnOffTypeModel(type='OnOff', value='OFF'))
        )


class PercentCommand:
    def percent(self: _HasNameAttr, value: float) -> None:
        """Command to value (in percent)"""
        if not 0 <= value <= 100:
            raise InvalidItemValueError.from_item(self, value)

        send_websocket_event(
            ItemCommandSendEvent.create(self.name, PercentTypeModel(type='Percent', value=str(value)))
        )


class UpDownCommand:

    def up(self: _HasNameAttr) -> None:
        """Command up"""
        send_websocket_event(
            ItemCommandSendEvent.create(self.name, UpDownTypeModel(type='UpDown', value='UP'))
        )

    def down(self: _HasNameAttr) -> None:
        """Command down"""
        send_websocket_event(
            ItemCommandSendEvent.create(self.name, UpDownTypeModel(type='UpDown', value='DOWN'))
        )

    def is_up(self: _HasNameAttr) -> bool:
        """Test value against on-value"""
        raise NotImplementedError()

    def is_down(self: _HasNameAttr) -> bool:
        """Test value against off-value"""
        raise NotImplementedError()



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Final

from HABApp.core.const import MISSING
from HABApp.openhab.connection.plugins import post_update, send_command, send_websocket_event
from HABApp.openhab.definitions.websockets import ItemCommandSendEvent, ItemStateSendEvent
from HABApp.openhab.definitions.websockets.item_value_types import RefreshTypeModel, StringTypeModel, UnDefTypeModel
from HABApp.openhab.item_to_reg import get_members
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent


if TYPE_CHECKING:
    Any = Any               # noqa: PLW0127
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/items/GroupItem.java
class GroupItem(OpenhabItem):
    """GroupItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar Any value: |oh_item_desc_value|
    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('GroupItem')
    _command_to_oh: Final = OutgoingCommandEvent('GroupItem')
    _state_from_oh_str = staticmethod(StringTypeModel.get_value_from_state)

    @property
    def members(self) -> tuple[OpenhabItem, ...]:
        """Resolves and then returns all group members"""

        return get_members(self.name)

    def oh_post_update(self, value: Any = MISSING) -> None:
        """Post an update to the openHAB item

        :param value: (optional) value to be posted. If not specified the current item value will be used.
        """
        new_value = self.value if value is MISSING else value

        if (obj := UnDefTypeModel.from_value(new_value)) is not None:
            send_websocket_event(ItemStateSendEvent.create(name=self._name, payload=obj))
            return None

        post_update(self._name, new_value, transport='http')
        return None

    def oh_send_command(self, value: Any = MISSING) -> None:
        """Send a command to the openHAB item

        :param value: (optional) value to be sent. If not specified the current item value will be used.
        """
        new_value = self.value if value is MISSING else value

        if (obj := RefreshTypeModel.from_value(new_value)) is not None:
            send_websocket_event(ItemCommandSendEvent.create(name=self._name, payload=obj))
            return None

        send_command(self._name, new_value, transport='http')
        return None



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Final

from eascheduler.job_control.base import BaseControl
from typing_extensions import Self, override
from whenever import Instant

from HABApp.core.asyncio import run_func_from_async
from HABApp.core.internals import uses_item_registry
from HABApp.core.items import BaseValueItem
from HABApp.openhab.items import OpenhabItem


if TYPE_CHECKING:
    from datetime import datetime as dt_datetime
    from datetime import timedelta as dt_timedelta

    from eascheduler.jobs import CountdownJob, DateTimeJob, OneTimeJob
    from eascheduler.jobs.base import JobBase


Items = uses_item_registry()


class HABAppBaseControl(BaseControl):
    def __init__(self) -> None:
        super().__init__()
        self._item: BaseValueItem | None = None

    def _cancel_timestamp_to_item(self, job: JobBase | None) -> None:
        self._timestamp_to_item(None)
        self._item = None
        self._job.on_update.remove(self._timestamp_to_item)
        self._job.on_finished.remove(self._cancel_timestamp_to_item)

    def _timestamp_to_item(self, job: JobBase | None) -> None:
        if self._item is None:
            return None

        if isinstance(self._item, OpenhabItem):
            self._item.oh_post_update(self.next_run_datetime)
        else:
            self._item.post_value(self.next_run_datetime)

    def to_item(self, item: str | BaseValueItem | None) -> None:
        """Sends the next execution (date)time to an item. Sends ``None`` if the job is not scheduled.
        Every time the scheduler updates to a new (date)time the item will also receive the updated time.

        :param item: item name or item, ``None`` to disable
        """
        if item is None:
            self._cancel_timestamp_to_item(None)
            return None

        self._item = Items.get_item(item if not isinstance(item, BaseValueItem) else item.name)
        self._job.on_update.register(self._timestamp_to_item)
        self._job.on_finished.register(self._cancel_timestamp_to_item)
        # Update the item with the current timestamp
        self._timestamp_to_item(None)

    @override
    def cancel(self: Self) -> Self:
        """Cancel the job"""

        run_func_from_async(self._job.job_finish)
        return self

    def get_next_run(self) -> dt_datetime | None:

        warnings.warn(
            'job.get_next_run() is deprecated. Use job.next_run_datetime',
            DeprecationWarning, stacklevel=2
        )
        return self.next_run_datetime

    def remaining(self) -> dt_timedelta | None:

        warnings.warn(
            'job.remaining() is deprecated. Use job.next_run_datetime to get the next execution time or None or'
            ' job.status to see if the job is running or not',
            DeprecationWarning, stacklevel=2
        )
        if (nr := self._job.next_run) is None:
            return None
        return (Instant.now() - nr).py_timedelta()


class CountdownJobControl(HABAppBaseControl):
    def __init__(self, job: CountdownJob) -> None:
        super().__init__()
        self._job: Final[CountdownJob] = job  # type: ignore[misc]

    def set_countdown(self, secs: float) -> Self:
        """Set the countdown time"""
        run_func_from_async(self._job.set_countdown, secs)
        return self

    def stop(self) -> Self:
        """Stop the countdown"""
        run_func_from_async(self._job.job_pause)
        return self

    def reset(self) -> Self:
        """Start the countdown again"""
        run_func_from_async(self._job.reset)
        return self


class OneTimeJobControl(HABAppBaseControl):
    def __init__(self, job: OneTimeJob) -> None:
        super().__init__()
        self._job: Final = job  # type: ignore[misc]


class DateTimeJobControl(HABAppBaseControl):
    def __init__(self, job: DateTimeJob) -> None:
        super().__init__()
        self._job: Final = job  # type: ignore[misc]

    def pause(self) -> Self:
        """Stop executing this job"""
        run_func_from_async(self._job.job_pause)
        return self

    def resume(self) -> Self:
        """Resume executing this job"""
        run_func_from_async(self._job.job_resume)
        return self



from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from HABApp.core.errors import InvalidItemValueError
from HABApp.openhab.definitions.websockets.item_value_types import OpenClosedTypeModel
from HABApp.openhab.interface_sync import post_update
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent


if TYPE_CHECKING:
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/ContactItem.java
class ContactItem(OpenhabItem):
    """ContactItem

    :ivar str name: |oh_item_desc_name|
    :ivar str value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('ContactItem', 'OpenClosed', 'UnDef')
    _command_to_oh: Final = OutgoingCommandEvent('ContactItem', 'Refresh')
    _state_from_oh_str = staticmethod(OpenClosedTypeModel.get_value_from_state)

    def set_value(self, new_value) -> bool:

        if new_value not in ('OPEN', 'CLOSED', None):
            raise InvalidItemValueError.from_item(self, new_value)

        return super().set_value(new_value)

    def is_open(self) -> bool:
        """Test value against open value"""
        return self.value == 'OPEN'

    def is_closed(self) -> bool:
        """Test value against closed value"""
        return self.value == 'CLOSED'

    def open(self) -> None:
        """Post an update to the item with the open value"""
        return post_update(self.name, 'OPEN')

    def closed(self) -> None:
        """Post an update to the item with the closed value"""
        return post_update(self.name, 'CLOSED')

    def __str__(self) -> str:
        return str(self.value)

    def __eq__(self, other):
        if isinstance(other, ContactItem):
            return self.value == other.value

        if isinstance(other, str):
            return self.value == other

        if isinstance(other, int):
            if other and self.is_open():
                return True
            if not other and self.is_closed():
                return True
            return False

        return NotImplemented



from .items import ItemResp, ItemRespList, ShortItemResp, ShortItemRespList
from .links import ItemChannelLinkResp, ItemChannelLinkRespList
from .persistence import ItemHistoryResp, PersistenceServiceResp, PersistenceServiceRespList
from .root import RootResp
from .systeminfo import SystemInfoRootResp
from .things import ChannelResp, ThingResp, ThingRespList
from .transformations import TransformationResp, TransformationRespList



from pydantic import BaseModel, TypeAdapter


# Documentation of TransformationDTO
# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.transform/src/main/java/org/openhab/core/io/rest/transform/TransformationDTO.java

class TransformationResp(BaseModel):
    uid: str
    label: str
    type: str
    configuration: dict[str, str]
    editable: bool


TransformationRespList = TypeAdapter(tuple[TransformationResp, ...])



from __future__ import annotations

import asyncio
import collections
import time
import typing
from datetime import timedelta

import HABApp
from HABApp.core.errors import ItemNameNotOfTypeStrError, ItemNotFoundException, WrongItemTypeError
from HABApp.core.events import EventFilter, ValueChangeEvent, ValueUpdateEvent
from HABApp.core.internals import (
    EventBusListener,
    uses_event_bus,
    uses_get_item,
    uses_item_registry,
    wrap_func,
)
from HABApp.core.items import BaseValueItem
from HABApp.core.wrapper import process_exception


get_item = uses_get_item()
item_registry = uses_item_registry()
event_bus = uses_event_bus()


class AggregationItem(BaseValueItem):

    @classmethod
    def get_create_item(cls, name: str) -> AggregationItem:
        """Creates a new AggregationItem in HABApp and returns it or returns the
        already existing item with the given name

        :param name: item name
        :return: item
        """
        if not isinstance(name, str):
            raise ItemNameNotOfTypeStrError.from_value(name)

        try:
            item = get_item(name)
        except ItemNotFoundException:
            item = cls(name)
            item_registry.add_item(item)

        if not isinstance(item, cls):
            raise WrongItemTypeError.from_item(item, cls)
        return item


    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.__period: float = 0
        self.__aggregation_func: typing.Callable[[typing.Iterable], typing.Any] = lambda x: x

        self._ts: collections.deque[float] = collections.deque()
        self._vals: collections.deque[typing.Any] = collections.deque()

        self.__listener: EventBusListener | None = None

        self.__task: asyncio.Future | None = None

    def aggregation_func(self, func: typing.Callable[[typing.Iterable], typing.Any]) -> AggregationItem:
        """Set the function which will be used to aggregate all values. E.g. ``min`` or ``max``

        :param func: The function which takes an iterator an returns an aggregated value.
                     Important: the function must be **non blocking**!
        """
        self.__aggregation_func = func
        return self

    def aggregation_period(self, period: float | int | timedelta) -> AggregationItem:
        """Set the period in which the items will be aggregated

        :param period: period in seconds
        """
        if isinstance(period, timedelta):
            period = period.total_seconds()

        assert period > 0, period
        self.__period = period

        # Clean old items (e.g. if we made the period shorter)
        while len(self._ts) > 1 and self._ts[1] + self.__period < time.time():
            self._ts.popleft()
            self._vals.popleft()

        return self

    def aggregation_source(self, source: BaseValueItem | str,
                           only_changes: bool = False) -> AggregationItem:
        """Set the source item which changes will be aggregated

        :param source: name or Item obj
        :param only_changes: if true only value changes instead of value updates will be added
        """

        # If we already have one we cancel it
        if self.__listener is not None:
            self.__listener.cancel()
            self.__listener = None

        self.__listener = EventBusListener(
            topic=source.name if isinstance(source, HABApp.core.items.BaseValueItem) else source,
            callback=wrap_func(self._add_value, name=f'{self.name}.add_value'),
            event_filter=EventFilter(ValueChangeEvent if only_changes else ValueUpdateEvent)
        )
        event_bus.add_listener(self.__listener)
        return self

    def _on_item_removed(self) -> None:
        super()._on_item_removed()

        if self.__listener is not None:
            self.__listener.cancel()
            self.__listener = None

        if self.__task is not None:
            self.__task.cancel()
            self.__task = None

    async def __update_task(self):
        try:
            while len(self._ts) > 1:
                ts = self._ts[1]
                now = time.time()

                left = (ts + self.__period) - now
                while left > 0:
                    await asyncio.sleep(left)
                    # sometimes we wake up to early
                    now = time.time()
                    left = (ts + self.__period) - now

                self._ts.popleft()
                self._vals.popleft()

                # old entries are removed -> now do the aggregation
                try:
                    val = self.__aggregation_func(self._vals)
                except Exception as e:
                    process_exception(self.__aggregation_func, e)
                    continue
                self.post_value(val)

        except Exception as e:
            process_exception(self.__update_task, e)
        finally:
            self.__task = None
        return None

    async def _add_value(self, event: ValueChangeEvent):
        self._ts.append(time.time())
        self._vals.append(event.value)

        if self.__task is None:
            self.__task = asyncio.create_task(self.__update_task())

        try:
            val = self.__aggregation_func(self._vals)
        except Exception as e:
            process_exception(self.__aggregation_func, e)
            return None

        self.post_value(val)
        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeAlias

import aiohttp

import HABApp
from HABApp.core.connections import AutoReconnectPlugin, BaseConnection, Connections, ConnectionStateToEventBusPlugin


if TYPE_CHECKING:
    from asyncio import Queue

    from HABApp.core.lib import InstantView
    from HABApp.openhab.definitions.websockets.base import BaseOutEvent
    from HABApp.openhab.items import OpenhabItem, Thing


@dataclass
class OpenhabContext:
    version: tuple[int, int, int]
    is_oh41: bool

    # true when we waited during connect
    waited_for_openhab: bool

    created_items: dict[str, tuple[OpenhabItem, InstantView]]
    created_things: dict[str, tuple[Thing, InstantView]]

    session: aiohttp.ClientSession
    session_options: dict[str, Any]

    out_queue: Queue[BaseOutEvent]

    @classmethod
    def new_context(cls, *, version: tuple[int, int, int],
                    session: aiohttp.ClientSession, session_options: dict[str, Any],
                    out_queue: Queue[BaseOutEvent]) -> OpenhabContext:
        return cls(
            version=version, is_oh41=version >= (4, 1),
            waited_for_openhab=False,
            created_items={}, created_things={},
            session=session, session_options=session_options,
            out_queue=out_queue
        )


CONTEXT_TYPE: TypeAlias = OpenhabContext | None


def setup() -> None:
    config = HABApp.config.CONFIG.openhab

    from HABApp.openhab.connection.handler import HANDLER as CONNECTION_HANDLER
    from HABApp.openhab.connection.plugins import (
        OUTGOING_PLUGIN,
        BrokenLinksPlugin,
        LoadOpenhabItemsPlugin,
        LoadTransformationsPlugin,
        PingPlugin,
        TextualThingConfigPlugin,
        ThingOverviewPlugin,
        WaitForPersistenceRestore,
        WaitForStartlevelPlugin,
        WebsocketPlugin,
    )

    connection = Connections.add(OpenhabConnection())
    connection.register_plugin(CONNECTION_HANDLER)

    connection.register_plugin(WaitForStartlevelPlugin(), 0)
    connection.register_plugin(OUTGOING_PLUGIN, 10)
    connection.register_plugin(LoadOpenhabItemsPlugin('LoadItemsAndThings'), 20)
    connection.register_plugin(WebsocketPlugin(), 30)
    connection.register_plugin(LoadOpenhabItemsPlugin('SyncItemsAndThings'), 40)
    connection.register_plugin(LoadTransformationsPlugin(), 50)
    connection.register_plugin(PingPlugin(), 100)
    connection.register_plugin(WaitForPersistenceRestore(), 110)
    connection.register_plugin(TextualThingConfigPlugin(), 120)
    connection.register_plugin(ThingOverviewPlugin(), 500_000)
    connection.register_plugin(BrokenLinksPlugin(), 500_001)

    connection.register_plugin(ConnectionStateToEventBusPlugin())
    connection.register_plugin(AutoReconnectPlugin())

    # config changes
    config.general.subscribe_for_changes(CONNECTION_HANDLER.update_cfg_general)


class OpenhabConnection(BaseConnection):
    def __init__(self) -> None:
        super().__init__('openhab')
        self.context: CONTEXT_TYPE = None

    def is_silent_exception(self, e: Exception) -> bool:
        from HABApp.openhab.connection.plugins.websockets import WebSocketClosedError

        return isinstance(e, (
            # https://docs.aiohttp.org/en/stable/client_reference.html#client-exceptions
            aiohttp.ClientError,

            # Websocket exceptions
            WebSocketClosedError,

            # aiohttp_sse_client Exceptions
            ConnectionRefusedError, ConnectionError, ConnectionAbortedError)
        )



from typing import Final, Literal


class RingCounterTracker:
    """Class that tracks a ring counter value and only allows increasing or decreasing values.
    """

    __slots__ = ('_ignore', '_is_increasing', '_last_value', '_max_value', '_min_value')

    def __init__(self, min_value: int | None = None, max_value: int | None = None, *,
                 ignore: int = 10, direction: Literal['increasing', 'decreasing'] = 'increasing') -> None:
        """

        :param min_value: Minimum value of the ring counter
        :param max_value: Maximum value of the ring counter
        :param ignore: How many values to ignore before allowing a lower value
        :param direction: Direction of the counter: increasing or decreasing
        """
        if max_value is None:
            max_value = min_value
            min_value = 0

        if not isinstance(min_value, int):
            msg = 'Min value must be an integer'
            raise TypeError(msg)
        if not isinstance(max_value, int):
            msg = 'Max value must be an integer'
            raise TypeError(msg)
        if not isinstance(ignore, int):
            msg = 'Ignore value must be an integer'
            raise TypeError(msg)

        if not min_value < max_value:
            msg = f'Min value {min_value} must be less than max value {max_value}'
            raise ValueError(msg)
        if not 0 < ignore < (max_value - min_value + 1):
            msg = f'Ignore value {ignore} must be greater than 0 and lower than max value {max_value}'
            raise ValueError(msg)

        self._min_value: Final = min_value
        self._max_value: Final = max_value
        self._ignore: Final = ignore
        self._is_increasing: Final = direction != 'decreasing'

        # # initialize so we always allow on first call
        self._last_value: int  = (max_value + ignore + 2) if self._is_increasing else (min_value - ignore - 2)

    @property
    def value(self) -> int:
        """Get the last value of the ring counter."""
        return self._last_value

    def __int__(self) -> int:
        return self._last_value

    def allow(self, value: int, *, strict: bool = True, set_value: bool = True) -> bool:
        """Return if a value is allowed and set it as the current value if it was allowed.

        :param value: Value to be checked
        :param strict: Check if the value is within min/max and of correct type
        :param set_value: Set the new value as the current value if it was allowed
        :return: True if the value was allowed, False if not
        """
        if strict and not isinstance(value, int):
            msg = f'Value must be an integer (is {type(value)})'
            raise TypeError(msg)

        if value > (max_value := self._max_value):
            if strict:
                msg = f'Value {value} is greater than max value {self._max_value}'
                raise ValueError(msg)
            return False

        if value < (min_value := self._min_value):
            if strict:
                msg = f'Value {value} is less than min value {self._min_value}'
                raise ValueError(msg)
            return False

        if self._is_increasing:
            lowest_ignored = (last_value := self._last_value) - self._ignore + 1
            if lowest_ignored <= value <= last_value:
                return False
            if lowest_ignored < min_value:
                lowest_ignored += (max_value - min_value + 1)
                if lowest_ignored <= value <= max_value:
                    return False
        else:
            highest_ignored = (last_value := self._last_value) + self._ignore - 1
            if last_value <= value <= highest_ignored:
                return False
            if highest_ignored > max_value:
                highest_ignored -= (max_value - min_value + 1)
                if min_value <= value <= highest_ignored:
                    return False

        if set_value:
            self._last_value = value
        return True

    def test_allow(self, value: int) -> bool:
        """Test if a value will be allowed without setting it as the current value.

        :param value: value to test
        :return: True if the value would be allowed, False if not
        """
        return self.allow(value, strict=False, set_value=False)

    def __repr__(self) -> str:
        value = self._last_value
        if value > self._max_value or value < self._min_value:
            value = '-'
        return (f'{self.__class__.__name__:s}(value={value}, '
                f'min={self._min_value:d}, max={self._max_value:d}, ignore={self._ignore:d})')

    def __eq__(self, other: int) -> bool:
        return self._last_value == other

    def __ne__(self, other: float) -> bool:
        return self._last_value != other

    def __ge__(self, other: float) -> bool:
        return self._last_value >= other

    def __gt__(self, other: float) -> bool:
        return self._last_value > other

    def __le__(self, other: float) -> bool:
        return self._last_value <= other

    def __lt__(self, other: float) -> bool:
        return self._last_value < other



import asyncio
import logging
import os
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any, TypeAlias

import HABApp
from HABApp.core.logger import HABAppError, HABAppWarning
from HABApp.core.wrapper import process_exception


log = logging.getLogger('HABApp.execute')

HINT_EXEC_ARGS: TypeAlias = str | Path
HINT_PYTHON_PATH: TypeAlias = Iterable[str | Path] | None


def _ensure_str_objs(objs: Iterable[HINT_EXEC_ARGS], key: str, enforce_abs=False) -> list[str]:
    new_args: list[str] = []

    # args must be str, but we support str and Path
    for i, val in enumerate(objs):
        if isinstance(val, Path):
            path_val = val
            str_val = str(val)
        elif isinstance(val, str):
            path_val = Path(val)
            str_val = val
        else:
            msg = f'{key:s}[{i:d}] is not of type str! "{val}" ({type(val).__name__:s})'
            raise TypeError(msg)

        if enforce_abs and not path_val.is_absolute():
            msg = f'{key:s}[{i:d}] is not an absolute path: "{val}"'
            raise ValueError(msg)

        new_args.append(str_val)

    return new_args


def build_exec_params(*args: HINT_EXEC_ARGS,
                      _capture_output: bool = True,
                      _additional_python_path: HINT_PYTHON_PATH = None,
                      **kwargs: Any) -> tuple[Iterable[str], dict[str, Any]]:
    # convenience for easy capturing
    if _capture_output:
        if 'stdout' in kwargs:
            msg = 'Parameter "capture_output" can not be used with "stdout" in kwargs!'
            raise ValueError(msg)
        kwargs['stdout'] = asyncio.subprocess.PIPE
        if 'stderr' in kwargs:
            msg = 'Parameter "capture_output" can not be used with "stderr" in kwargs!'
            raise ValueError(msg)
        kwargs['stderr'] = asyncio.subprocess.PIPE

    # convenience for additional libraries
    if _additional_python_path is not None:
        if 'env' in kwargs:
            msg = 'Parameter "additional_python_path" can not be used with "env" in kwargs!'
            raise ValueError(msg)

        ppath = _ensure_str_objs(_additional_python_path, 'additional_python_path', enforce_abs=True)

        # The child process needs the env from the current process, that's why we merge them
        # See docs of subprocess.Popen -> env for more details
        env = dict(os.environ)

        existing_ppath = env.get('PYTHONPATH')
        if existing_ppath:
            ppath.append(existing_ppath)
        env['PYTHONPATH'] = os.pathsep.join(ppath)
        kwargs['env'] = env

    # set config folder as working directory
    if 'cwd' not in kwargs:
        kwargs['cwd'] = HABApp.CONFIG._file_path.parent

    return _ensure_str_objs(args, 'args'), kwargs


class FinishedProcessInfo:
    """Information about the finished process.

    :var int returncode: Return code of the process
    :var Optional[str] stdout: Standard output of the process or ``None``
    :var Optional[str] stderr: Error output of the process or ``None``
    """

    def __init__(self, returncode: int, stdout: str | None, stderr: str | None) -> None:
        self.returncode: int = returncode
        self.stdout: str | None = stdout
        self.stderr: str | None = stderr

    def __repr__(self) -> str:
        return f'<ProcessInfo: returncode: {self.returncode}, stdout: {self.stdout}, stderr: {self.stderr}>'

    def __eq__(self, other):
        if isinstance(other, FinishedProcessInfo):
            return self.returncode == other.returncode and self.stdout == other.stdout and self.stderr == other.stderr

        return NotImplementedError


HINT_PROCESS_CB_FULL: TypeAlias = Callable[[FinishedProcessInfo], Any]
HINT_PROCESS_CB_SIMPLE: TypeAlias = Callable[[str], Any]


async def async_subprocess_exec(callback, *args, calling_func, raw_info: bool, **kwargs):
    call_str = ''

    try:

        proc = None
        stdout = None
        stderr = None

        call_str = ' '.join(f'"{x}"' for x in args)

        try:
            proc = await asyncio.create_subprocess_exec(*args, **kwargs)

            b_stdout, b_stderr = await proc.communicate()
            ret_code = proc.returncode

            if b_stdout or b_stderr:
                stdout = b_stdout.decode()
                stderr = b_stderr.decode()

        except asyncio.CancelledError:
            if proc is not None:
                proc.terminate()

            HABAppWarning(log=log).add(f'Subprocess canceled! Call: {call_str}').dump()
            return None

    except FileNotFoundError as e:
        HABAppError(log=log).add('Creating subprocess failed!').add(f'  Call: {call_str}').add(
            f'  Working dir: {kwargs.get("cwd")}').add(str(e)).dump()
        return None
    except Exception as e:
        HABAppError(log=log).add('Creating subprocess failed!').add(f'  Call: {call_str}').add(
            f'  Working dir: {kwargs.get("cwd")}').dump()
        process_exception(calling_func, e, logger=log)
        return None

    if raw_info:
        # callback is a wrapped function, that's why it will not throw an error
        callback(FinishedProcessInfo(ret_code, stdout, stderr))
        return None

    if ret_code != 0:
        e = HABAppError(log=log).add(f'Process returned {ret_code:d}!').add(f'Call: {call_str}')
        if stderr:
            e.add('Stderr:')
            for line in stderr.splitlines():
                e.add(f'  {line}')
        elif stdout:
            e.add('Stdout:')
            for line in stdout.splitlines():
                e.add(f'  {line}')
        e.dump()
        return None

    # no error -> callback always with str
    callback(stdout if stdout else '')
    return None



from typing import Final

from typing_extensions import Self


class RingCounter:
    """A ring counter is a counter that wraps around when it reaches its maximum value.
    """

    __slots__ = ('_max_value', '_min_value', '_value')

    def __init__(self, min_value: int | None = None, max_value: int | None = None, *,
                 initial_value: int | None = None) -> None:
        if max_value is None:
            max_value = min_value
            min_value = 0

        if initial_value is None:
            initial_value = min_value

        if not isinstance(min_value, int):
            msg = 'Min value must be an integer'
            raise TypeError(msg)
        if not isinstance(max_value, int):
            msg = 'Max value must be an integer'
            raise TypeError(msg)
        if not isinstance(initial_value, int):
            msg = 'Initial value must be an integer'
            raise TypeError(msg)

        if not min_value < max_value:
            msg = f'Min value {min_value} must be less than max value {max_value}'
            raise ValueError(msg)
        if not min_value <= initial_value <= max_value:
            msg = f'Initial value {initial_value} is not in range [{min_value:d}..{max_value:d}]'
            raise ValueError(msg)

        self._min_value: Final = min_value
        self._max_value: Final = max_value
        self._value: int = initial_value

    @property
    def size(self) -> int:
        """Return the size (how man values it can count) of the ring counter."""
        return self._max_value - self._min_value + 1

    def __len__(self) -> int:
        return self.size

    @property
    def value(self) -> int:
        """Current value of the ring counter."""
        return self._value

    def __int__(self) -> int:
        return self._value

    def increase(self, value: int = 1) -> Self:
        """Increase the value of the ring counter by the given value.

        :param value: How much to increase the value by.
        """
        if value < 0:
            msg = 'Value must be >= 0'
            raise ValueError(msg)

        self._value += value

        while self._value > self._max_value:
            self._value -= self.size

        return self

    def decrease(self, value: int = 1) -> Self:
        """Decrease the value of the ring counter by the given value.

        :param value: How much to decrease the value by.
        """
        if value < 0:
            msg = 'Value must be >= 0'
            raise ValueError(msg)

        self._value -= value

        while self._value < self._min_value:
            self._value += self.size

        return self

    def __iadd__(self, other: int) -> Self:
        return self.increase(other)

    def __isub__(self, other: int) -> Self:
        return self.decrease(other)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(value={self._value:d}, min={self._min_value:d} max={self._max_value:d})'

    def __eq__(self, other: int) -> bool:
        return self._value == other

    def __ne__(self, other: float) -> bool:
        return self._value != other

    def __ge__(self, other: float) -> bool:
        return self._value >= other

    def __gt__(self, other: float) -> bool:
        return self._value > other

    def __le__(self, other: float) -> bool:
        return self._value <= other

    def __lt__(self, other: float) -> bool:
        return self._value < other



import logging
from asyncio import iscoroutinefunction
from collections.abc import Callable, Coroutine
from typing import Any

from HABApp.config import CONFIG
from HABApp.core import shutdown
from HABApp.core.internals import Context
from HABApp.core.internals.wrapped_function.base import P, R, WrappedFunctionBase
from HABApp.core.internals.wrapped_function.wrapped_async import WrappedAsyncFunction
from HABApp.core.internals.wrapped_function.wrapped_sync import WrappedSyncFunction
from HABApp.core.internals.wrapped_function.wrapped_thread import (
    WrappedThreadFunction,
    create_thread_pool,
    stop_thread_pool,
)


def wrap_func(func: Callable[P, R] | Callable[P, Coroutine[Any, Any, R]],
              warn_too_long=True,
              name: str | None = None,
              logger: logging.Logger | None = None,
              context: Context | None = None) -> WrappedFunctionBase[P, R]:

    # Check that it's actually a callable, so we fail fast and not when we try to run the function.
    # Some users pass the result of the function call (e.g. func()) by accident
    # which will inevitably fail once we try to run the function.
    if not callable(func):
        try:
            type_name: str = func.__class__.__name__
        except Exception:
            type_name = type(func)

        msg = f'Callable or coroutine function expected! Got "{func}" (type {type_name:s})'
        raise TypeError(msg)

    if iscoroutinefunction(func):
        return WrappedAsyncFunction(func, name=name, logger=logger, context=context)

    return SYNC_CLS(func, warn_too_long=warn_too_long, name=name, logger=logger, context=context)


SYNC_CLS: type[WrappedThreadFunction] | type[WrappedSyncFunction]


def setup() -> None:
    global SYNC_CLS

    if not THREAD_POOL.enabled:
        SYNC_CLS = WrappedSyncFunction

        # In case of hot reload
        stop_thread_pool()
        return None

    SYNC_CLS = WrappedThreadFunction

    # create thread pool
    create_thread_pool(THREAD_POOL.threads)

    # this function can be called multiple times, so it's no problem if we register it more than once!
    shutdown.register(stop_thread_pool, msg='Stopping thread pool', last=True)


THREAD_POOL = CONFIG.habapp.thread_pool
THREAD_POOL.subscribe_for_changes(setup)



from typing import Any

from pydantic import BaseModel, Field, TypeAdapter


class ItemChannelLinkResp(BaseModel):
    # AbstractLinkDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/link/dto/AbstractLinkDTO.java
    item: str = Field(alias='itemName')

    # ItemChannelLinkDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/link/dto/ItemChannelLinkDTO.java
    channel: str = Field(alias='channelUID')
    configuration: dict[str, Any] = {}

    # EnrichedItemChannelLinkDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/link/EnrichedItemChannelLinkDTO.java
    editable: bool


ItemChannelLinkRespList = TypeAdapter(tuple[ItemChannelLinkResp, ...])



import ast
import datetime
import importlib
from collections.abc import Callable
from inspect import isclass, ismodule
from pathlib import Path
from typing import Any, Final

import whenever
from easyconfig.config_objs import ConfigObj
from immutables import Map
from stack_data import Variable

from HABApp.core.const.json import dump_json, load_json

from .const import PRE_INDENT, SEPARATOR_VARIABLES


# don't show these types in the traceback
SKIPPED_TYPES = (
    bool, bytearray, bytes, complex, dict, float, frozenset, int, list, memoryview, set, str, tuple, type(None),
    datetime.date, datetime.datetime, datetime.time, datetime.timedelta,
    Map, Path,
    whenever.Instant,
    whenever.SystemDateTime, whenever.PlainDateTime, whenever.ZonedDateTime, whenever.OffsetDateTime,
    whenever.TimeDelta, whenever.DateDelta, whenever.DateTimeDelta
)


def is_type_hint_or_type(value: Any) -> bool:
    if isinstance(value, tuple):
        return all(is_type_hint_or_type(obj) for obj in value)

    # compare through identity since some objects override the equality operator.
    # If we check with "value in SKIPPED_TYPES" we might get an exception because there the __eq__ operator is used
    if any(value is o for o in SKIPPED_TYPES):
        return True

    # check if it's something from the typing module
    # we can't do that with isinstance, so we try this
    try:
        if value.__module__ == 'typing' or value.__class__.__module__ == 'typing':
            return True
    except AttributeError:
        pass

    return False


def _filter_expressions(name: str, value: Any) -> bool:
    # a is None = True
    if name.endswith(' is None'):
        return True

    return False


SKIPPED_OBJS: Final[tuple[str, ...]] = (
    'HABApp.core.Items',
)


def _skip_objs(name: str, value: Any) -> bool:
    for dotted_path in SKIPPED_OBJS:
        path = dotted_path.split('.')
        obj = importlib.import_module('.'.join(path[:-1]))
        if value is getattr(obj, path[-1]):
            return True
    return False


SKIP_VARIABLE: tuple[Callable[[str, Any], bool], ...] = (
    # module imports
    lambda name, value: ismodule(value),

    # type hints and type tuples
    lambda name, value: is_type_hint_or_type(value),

    # functions
    lambda name, value: value is dump_json or value is load_json,

    # config value objs
    lambda name, value: isinstance(value, ConfigObj),

    # Expressions
    _filter_expressions,
)

ORDER_VARIABLE: tuple[Callable[[Variable], bool], ...] = (
    lambda x: isclass(x.value),
)


def skip_variable(var: Variable) -> bool:
    name = var.name
    value = var.value
    return any(func(name, value) for func in SKIP_VARIABLE)


def format_frame_variables(tb: list[str], stack_variables: list[Variable]):
    if not stack_variables:
        return None

    # remove variables that shall not be printed
    used_vars: set[Variable] = {v for v in stack_variables if not skip_variable(v)}

    # remove objs and attributes
    rem_obj_names = {v.name for v in used_vars if _skip_objs(v.name, v.value)}
    rem_objs = {v for v in used_vars for rem_name in rem_obj_names if v.name.startswith(rem_name)}
    used_vars -= rem_objs

    # attributes
    dotted_names: set[str] = {n.name.split('.')[0] for n in used_vars if '.' in n.name}

    # Sort output
    used_vars = sorted(used_vars, key=lambda x: (
        isinstance(x.nodes[0], ast.Compare),                                        # Compare objects last
        not any(map(
            lambda y: x.name == y or x.name.startswith(y + '.'), dotted_names)),    # Classes with attributes
        x.name.lower()                                                              # Name lowercase
    ))

    variables = {}

    # variables by order
    for add_var in ORDER_VARIABLE:
        for var in used_vars:
            name = var.name
            if name in variables:
                continue
            if add_var(var):
                variables[name] = var.value

    # rest of the variables
    for var in used_vars:
        if var.name not in variables:
            variables[var.name] = var.value

    if not variables:
        return None

    # Add variables to traceback
    tb.append(SEPARATOR_VARIABLES)

    for name, value in variables.items():
        # both name and value can be a multiline string
        # -> try to format it nicely

        name_lines = name.splitlines()
        for line in name_lines[:-1]:
            tb.append(f'{" " * (PRE_INDENT + 1):s}{line:s}')

        last_name_line = name_lines[-1]
        for nr, line in enumerate(repr(value).splitlines()):
            if not nr:
                tb.append(f'{" " * (PRE_INDENT + 1):s}{last_name_line:s} = {line}')
            else:
                tb.append(f'{" " * (PRE_INDENT + 1):s}{" " * len(last_name_line):s}   {line}')

    tb.append(SEPARATOR_VARIABLES)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from immutables import Map

import HABApp
from HABApp.core.internals import uses_item_registry
from HABApp.core.logger import log_warning
from HABApp.openhab.definitions.things import (
    THING_STATUS_DEFAULT,
    THING_STATUS_DETAIL_DEFAULT,
    ThingStatusDetailEnum,
    ThingStatusEnum,
)


if TYPE_CHECKING:
    from HABApp.openhab.definitions.rest import ThingResp
    from HABApp.openhab.events import ThingAddedEvent
    from HABApp.openhab.items import OpenhabItem

log = logging.getLogger('HABApp.openhab.items')

Items = uses_item_registry()


# noinspection PyProtectedMember
def add_to_registry(item: OpenhabItem, *, set_value: bool = False) -> None:
    name = item.name
    for grp in item.groups:
        MEMBERS.setdefault(grp, set()).add(name)

    if not Items.item_exists(name):
        Items.add_item(item)
        return None

    existing = Items.get_item(name)
    if isinstance(existing, item.__class__):
        # If we load directly through the API and not through an event we have to set the value
        if set_value:
            existing.set_value(item.value)

        # remove old groups
        for grp in set(existing.groups) - set(item.groups):
            MEMBERS.get(grp, set()).discard(name)

        # same type - it was only an item update (e.g. label)!
        existing._update_item_definition(item)
        return None

    log_warning(log, f'Item type changed from {existing.__class__} to {item.__class__}')

    # Replace existing item with the updated definition
    Items.pop_item(name)
    Items.add_item(item)
    return None


def remove_from_registry(name: str) -> None:
    if not Items.item_exists(name):
        return None

    item = Items.get_item(name)  # type: HABApp.openhab.items.OpenhabItem
    for grp in item.groups:
        m = MEMBERS.get(grp, set())
        m.discard(name)
        if not m:
            MEMBERS.pop(grp, None)

    Items.pop_item(name)
    return None


MEMBERS: dict[str, set[str]] = {}


def fresh_item_sync() -> None:
    MEMBERS.clear()


def get_members(group_name: str) -> tuple[OpenhabItem, ...]:
    ret = []
    for name in MEMBERS.get(group_name, ()):
        item = Items.get_item(name)  # type: OpenhabItem
        ret.append(item)
    return tuple(sorted(ret, key=lambda x: x.name))


# ----------------------------------------------------------------------------------------------------------------------
# Thing handling
# ----------------------------------------------------------------------------------------------------------------------

def get_thing_status_from_resp(
        obj: ThingResp | None) -> tuple[ThingStatusEnum, ThingStatusDetailEnum, str]:
    if obj is None:
        return THING_STATUS_DEFAULT, THING_STATUS_DETAIL_DEFAULT, ''
    return (
        obj.status.status,
        obj.status.detail,
        obj.status.description if obj.status.description is not None else ''
    )


def add_thing_to_registry(data: ThingResp | ThingAddedEvent):

    if isinstance(data, HABApp.openhab.events.thing_events.ThingAddedEvent):
        name = data.name
        status, status_detail, status_description = get_thing_status_from_resp(None)
    elif isinstance(data, HABApp.openhab.definitions.rest.ThingResp):
        name = data.uid
        status, status_detail, status_description = get_thing_status_from_resp(data)
    else:
        raise TypeError()

    if Items.item_exists(name):
        existing = Items.get_item(name)
        if isinstance(existing, HABApp.openhab.items.Thing):
            new_thing = existing
        else:
            # Replace existing item with the correct type
            new_thing = HABApp.openhab.items.Thing(name=name)
            log_warning(log, f'Item type changed from {existing.__class__} to {new_thing.__class__}')
            Items.pop_item(name)
    else:
        new_thing = HABApp.openhab.items.Thing(name=name)

    new_thing.status = status
    new_thing.status_detail = status_detail
    new_thing.status_description = status_description
    new_thing.label = data.label
    new_thing.location = data.location
    new_thing.configuration = Map(data.configuration)
    new_thing.properties = Map(data.properties)
    return Items.add_item(new_thing)


def remove_thing_from_registry(name: str):
    if not Items.item_exists(name):
        return None
    return Items.pop_item(name)



from typing import Any, Final

import HABApp.core

from .base_event import OpenhabEvent


class ItemStateEvent(OpenhabEvent, HABApp.core.events.ValueUpdateEvent):

    def __repr__(self) -> str:
        value = f() if (f := getattr(value := self.value, '_value_str', None)) is not None else str(value)
        return f'<{self.__class__.__name__} name: {self.name}, value: {value}>'


class ItemStateUpdatedEvent(OpenhabEvent, HABApp.core.events.ValueUpdateEvent):

    def __repr__(self) -> str:
        value = f() if (f := getattr(value := self.value, '_value_str', None)) is not None else str(value)
        return f'<{self.__class__.__name__} name: {self.name}, value: {value}>'


class ItemStateChangedEvent(OpenhabEvent, HABApp.core.events.ValueChangeEvent):

    def __repr__(self) -> str:
        value = f() if (f := getattr(value := self.value, '_value_str', None)) is not None else str(value)
        old_value = f() if (f := getattr(old_value := self.old_value, '_value_str', None)) is not None \
            else str(old_value)
        return f'<{self.__class__.__name__} name: {self.name}, value: {value}, old_value: {old_value}>'


class ItemCommandEvent(OpenhabEvent, HABApp.core.events.ValueCommandEvent):

    def __repr__(self) -> str:
        value = f() if (f := getattr(value := self.value, '_value_str', None)) is not None else str(value)
        return f'<{self.__class__.__name__} name: {self.name}, value: {value}>'


class ItemAddedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str type:
    :ivar str | None label:
    :ivar frozenset[str] tags:
    :ivar frozenset[str] groups:
    """
    name: str
    type: str
    label: str | None
    tags: frozenset[str]
    groups: frozenset[str]

    def __init__(self, name: str, type: str, label: str | None,
                 tags: frozenset[str], group_names: frozenset[str]) -> None:
        super().__init__()

        self.name: str = name
        self.type: str = type
        self.label: str | None = label
        self.tags: frozenset[str] = tags
        self.groups: frozenset[str] = group_names

    def __repr__(self) -> str:
        tags = f' {{{", ".join(sorted(self.tags))}}}' if self.tags else ''
        grps = f' {{{", ".join(sorted(self.groups))}}}' if self.groups else ''
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}, tags:{tags}, groups:{grps}>'


class ItemUpdatedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str type:
    :ivar str | None label:
    :ivar frozenset[str] tags:
    :ivar frozenset[str] groups:
    """
    name: str
    type: str
    label: str | None
    tags: frozenset[str]
    groups: frozenset[str]

    def __init__(self, name: str, type: str, label: str | None,
                 tags: frozenset[str], group_names: frozenset[str]) -> None:
        super().__init__()

        self.name: str = name
        self.type: str = type
        self.label: str | None = label
        self.tags: frozenset[str] = tags
        self.groups: frozenset[str] = group_names

    def __repr__(self) -> str:
        tags = f' {{{", ".join(sorted(self.tags))}}}' if self.tags else ''
        grps = f' {{{", ".join(sorted(self.groups))}}}' if self.groups else ''
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}, tags:{tags}, groups:{grps}>'


class ItemRemovedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str type:
    :ivar str | None label:
    :ivar frozenset[str] tags:
    :ivar frozenset[str] groups:
    """
    name: str
    type: str
    label: str | None
    tags: frozenset[str]
    groups: frozenset[str]

    def __init__(self, name: str, type: str, label: str | None,
                 tags: frozenset[str], groups: frozenset[str]) -> None:
        super().__init__()

        self.name: str = name
        self.type: str = type
        self.label: str | None = label
        self.tags: frozenset[str] = tags
        self.groups: frozenset[str] = groups

    def __repr__(self) -> str:
        tags = f' {{{", ".join(sorted(self.tags))}}}' if self.tags else ''
        grps = f' {{{", ".join(sorted(self.groups))}}}' if self.groups else ''
        return f'<{self.__class__.__name__} name: {self.name}, type: {self.type}, tags:{tags}, groups:{grps}>'


class ItemStatePredictedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar Any value:
    :ivar bool is_confirmation:
    """
    name: str
    value: Any
    is_confirmation: bool

    def __init__(self, name: str, value: Any, is_confirmation: bool) -> None:  # noqa: FBT001
        super().__init__()
        self.name: Final = name
        self.value: Final = value
        self.is_confirmation: Final = is_confirmation

    def __repr__(self) -> str:
        return (f'<{self.__class__.__name__} name: {self.name}, value: {self.value} '
                f'is_confirmation: {self.is_confirmation}>')


class GroupStateUpdatedEvent(OpenhabEvent, HABApp.core.events.ValueUpdateEvent):
    """
    :ivar str name: Group name
    :ivar str item: Group item that caused the update
    :ivar Any value:
    """
    name: str
    item: str
    value: Any

    def __init__(self, name: str, item: str, value: Any) -> None:
        super().__init__(name, value)
        self.item: Final = item

    def __repr__(self) -> str:
        value = f() if (f := getattr(value := self.value, '_value_str', None)) is not None else str(value)
        return f'<{self.__class__.__name__} name: {self.name}, item: {self.item}, value: {value}>'


class GroupStateChangedEvent(OpenhabEvent, HABApp.core.events.ValueChangeEvent):
    """
    :ivar str name:
    :ivar str item:
    :ivar Any value:
    :ivar Any old_value:
    """
    name: str
    item: str
    value: Any
    old_value: Any

    def __init__(self, name: str, item: str, value: Any, old_value: Any) -> None:
        super().__init__(name, value, old_value)
        self.item: Final = item

    def __repr__(self) -> str:
        value = f() if (f := getattr(value := self.value, '_value_str', None)) is not None else str(value)
        old_value = f() if (f := getattr(old_value := self.old_value, '_value_str', None)) is not None \
            else str(old_value)
        return (f'<{self.__class__.__name__} name: {self.name}, '
                f'item: {self.item}, value: {value}, old_value: {old_value}>')



from pathlib import Path


def _get_file_name(file: Path, nr: int) -> Path:
    if nr < 0:
        raise ValueError()

    if nr:
        return file.with_name(f'{file.name:s}.{nr:d}')
    return file


def rotate_file(file: Path, backup_count: int) -> None:
    if not isinstance(backup_count, int):
        raise TypeError()

    _get_file_name(file, backup_count).unlink(missing_ok=True)
    for i in range(backup_count - 1, -1, -1):
        src = _get_file_name(file, i)
        if not src.is_file():
            continue

        dst = _get_file_name(file, i + 1)
        dst.unlink(missing_ok=True)
        src.rename(dst)



from collections.abc import Callable


def get_obj_name(obj: Callable) -> str:
    # not all callables have a __name__ attribute, i.e. Mock, lambda
    if (name := getattr(obj, '__name__', None)) is None:
        return repr(obj)
    return name



from typing import Any, Final


class ValueUpdateEvent:
    """
    :ivar str name:
    :ivar Any value:
    """

    name: str
    value: Any

    def __init__(self, name: str, value: Any) -> None:
        self.name: Final = name
        self.value: Final = value

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} name: {self.name:s}, value: {self.value}>'


class ValueChangeEvent:
    """
    :ivar str name:
    :ivar Any value:
    :ivar Any old_value:
    """

    name: str
    value: Any
    old_value: Any

    def __init__(self, name: str, value: Any, old_value: Any) -> None:
        self.name: Final = name
        self.value: Final = value
        self.old_value: Final = old_value

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} name: {self.name:s}, value: {self.value}, old_value: {self.old_value}>'


class ValueCommandEvent:
    """
    :ivar str name:
    :ivar Any value:
    """

    name: str
    value: Any

    def __init__(self, name: str, value: Any) -> None:
        self.name: Final = name
        self.value: Final = value

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} name: {self.name:s}, value: {self.value}>'


class ItemNoChangeEvent:
    """
    :ivar str name:
    :ivar int | float seconds:
    """

    name: str
    seconds: int | float

    def __init__(self, name: str, seconds: int | float) -> None:
        self.name: Final = name
        self.seconds: Final = seconds

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} name: {self.name:s}, seconds: {self.seconds}>'


class ItemNoUpdateEvent:
    """
    :ivar str               name:
    :ivar int | float seconds:
    """
    name: str
    seconds: int | float

    def __init__(self, name: str, seconds: int | float) -> None:
        self.name: Final = name
        self.seconds: Final = seconds

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} name: {self.name:s}, seconds: {self.seconds}>'



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from threading import Lock
from time import monotonic
from typing import TYPE_CHECKING, Any, Final

from typing_extensions import override

from HABApp.core.asyncio import run_func_from_async, thread_context
from HABApp.core.const import loop
from HABApp.core.internals import Context, ContextProvidingObj
from HABApp.core.internals.wrapped_function.base import P, R, WrappedFunctionBase, default_logger


if TYPE_CHECKING:
    import logging
    from collections.abc import Callable


POOL: ThreadPoolExecutor | None = None
POOL_THREADS: int = 0


def _initialize_thread() -> None:
    thread_context.set('HABAppWorker')


def create_thread_pool(count: int) -> None:
    global POOL, POOL_THREADS

    if not isinstance(count, int) or count <= 0:
        msg = 'Thread count must be a positive integer'
        raise ValueError(msg)

    default_logger.debug(f'Starting thread pool with {count:d} threads!')

    stop_thread_pool()
    POOL_THREADS = count
    POOL = ThreadPoolExecutor(count, 'HabAppWorker', initializer=_initialize_thread)


def stop_thread_pool() -> None:
    global POOL, POOL_THREADS

    if (pool := POOL) is None:
        return None

    POOL_THREADS = 0
    POOL = None

    pool.shutdown()
    default_logger.debug('Thread pool stopped!')


POOL_INFO: Final[set[PoolFunc]] = set()
POOL_LOCK: Final = Lock()


class PoolFunc(ContextProvidingObj):
    def __init__(self, parent: WrappedThreadFunction, func_obj: Callable[..., Any], func_args: tuple[Any, ...],
                 func_kwargs: dict[str, Any], context: Context | None = None, **kwargs: Any) -> None:
        super().__init__(context=context, **kwargs)
        self.parent: Final = parent
        self.func_obj: Final = func_obj
        self.func_args: Final = func_args
        self.func_kwargs: Final = func_kwargs

        # timing checks
        self.submitted: float = monotonic()
        self.dur_start: float = 0.0
        self.dur_run: float = 0.0

        # thread info
        self.usage_high: int = 0

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} high: {self.usage_high:d}/{POOL_THREADS:d}>'

    def run(self) -> Any:
        pool_lock: Final = POOL_LOCK
        pool_info: Final = POOL_INFO
        parent = self.parent

        try:
            ts_start = monotonic()
            self.dur_start = ts_start - self.submitted

            with pool_lock:
                pool_info.add(self)
                count = len(pool_info)
                for info in pool_info:
                    info.usage_high = max(count, info.usage_high)

            # notify if we don't process quickly
            if self.dur_start > 0.05:
                parent.log.warning(f'Starting of {parent.name} took too long: {self.dur_start:.2f}s. '
                                   f'Maybe there are not enough threads?')

            # Profiler does not work
            # https://github.com/python/cpython/issues/124656

            # Execute the function
            ret = self.func_obj(*self.func_args, **self.func_kwargs)

            # log warning if execution takes too long
            self.dur_run = monotonic() - ts_start

            if parent.warn_too_long and self.dur_run > 0.8 and self.usage_high >= POOL_THREADS * 0.6:
                parent.log.warning(f'{self.usage_high}/{POOL_THREADS} threads have been in use and '
                                   f'execution of {parent.name} took too long: {self.dur_run:.2f}s')

        except Exception as e:
            # Process and dump the exception traceback from a coroutine.
            # That way we effectively serialize the logged tracebacks in case two exceptions happen at once
            run_func_from_async(self.parent.process_exception, e, *self.func_args, **self.func_kwargs)
            return None
        else:
            return ret
        finally:
            with pool_lock:
                pool_info.discard(self)


class WrappedThreadFunction(WrappedFunctionBase[P, R]):

    def __init__(self, func: Callable[P, R],
                 warn_too_long: bool = True,
                 name: str | None = None,
                 logger: logging.Logger | None = None,
                 context: Context | None = None) -> None:

        super().__init__(name=name, func=func, logger=logger, context=context)

        self.func: Final = func
        self.warn_too_long: Final = warn_too_long

    @override
    def run(self, *args: P.args, **kwargs: P.kwargs) -> None:
        # we need to copy the context, so it's available when the function is run
        pool_func = PoolFunc(self, self.func, args, kwargs, context=self._habapp_ctx)
        POOL.submit(pool_func.run)
        return None

    @override
    async def async_run(self, *args: P.args, **kwargs: P.kwargs) -> R | None:

        pool_func = PoolFunc(self, self.func, args, kwargs, context=self._habapp_ctx)
        return await loop.run_in_executor(POOL, pool_func.run)



from __future__ import annotations

from typing_extensions import Self


class QuantityInt(int):
    """Int which additionally carries a unit"""

    unit: str   # unit as sent from openHAB

    def __new__(cls, v: int, unit: str) -> Self:
        _obj = super().__new__(cls, v)
        _obj.unit = unit
        return _obj

    def _value_str(self) -> str:
        if unit := self.unit:
            return f'{super().__repr__()} {unit}'
        return super().__repr__()


class QuantityFloat(float):
    """Float which additionally carries a unit"""

    unit: str   # unit as sent from openHAB

    def __new__(cls, v: float, unit: str) -> Self:
        _obj = super().__new__(cls, v)
        _obj.unit = unit
        return _obj

    def _value_str(self) -> str:
        if unit := self.unit:
            return f'{super().__repr__()} {unit}'
        return super().__repr__()



from pydantic import BaseModel, Field, TypeAdapter


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.persistence/src/main/java/org/openhab/core/persistence/dto/PersistenceServiceDTO.java
class PersistenceServiceResp(BaseModel):
    id: str
    label: str | None = None
    type: str | None = None


PersistenceServiceRespList = TypeAdapter(tuple[PersistenceServiceResp, ...])


class DataPoint(BaseModel):
    time: int
    state: str


# ItemHistoryDTO
# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.persistence/src/main/java/org/openhab/core/persistence/dto/ItemHistoryDTO.java

class ItemHistoryResp(BaseModel):
    name: str
    total_records: str | None = Field(default=None, alias='totalrecords')
    data_points: str | None = Field(default=None, alias='datapoints')
    data: list[DataPoint] = []



from enum import Enum

from easyconfig.models import BaseModel, Field
from pydantic import AnyHttpUrl, ByteSize, TypeAdapter, field_validator


class Ping(BaseModel):
    enabled: bool = Field(True, description='If enabled the configured item will show how long it takes to send '
                                            'an update from HABApp and get the updated value back from openHAB '
                                            'in milliseconds')
    item: str = Field('HABApp_Ping', description='Name of the Numberitem')
    interval: int | float = Field(10, description='Seconds between two pings', ge=0.1)


class General(BaseModel):
    listen_only: bool = Field(
        False, description='If True HABApp does not change anything on the openHAB instance.'
    )
    wait_for_openhab: bool = Field(
        True,
        description='If True HABApp will wait for a successful openHAB connection before loading any rules on startup'
    )

    # Advanced settings
    min_start_level: int = Field(
        70, ge=0, le=100, in_file=False,
        description='Minimum openHAB start level to load items and listen to events',
    )

    # Minimum uptime
    min_uptime: int = Field(
        60, ge=0, le=3600 * 2, in_file=False,
        description='Minimum openHAB uptime in seconds to load items and listen to events',
    )


class EventTypeFilterEnum(str, Enum):
    OFF = 'OFF'
    AUTO = 'AUTO'
    CONFIG = 'CONFIG'

    def is_auto(self):
        return self == EventTypeFilterEnum.AUTO

    def is_config(self):
        return self == EventTypeFilterEnum.CONFIG


class WebsocketEventFilter(BaseModel):
    event_type: EventTypeFilterEnum = Field(
        EventTypeFilterEnum.AUTO, alias='type',
        description='Configure the event type filter. "OFF" to allow all events, "AUTO" to allow the recommended '
                    'events or "CONFIG" to use the event types from the "types allowed" field.'
    )
    types_allowed: tuple[str, ...] = Field(
        (), alias='types allowed', description='List of event types that will be allowed'
    )


class Websocket(BaseModel):
    max_msg_size: ByteSize = Field(
        '4Mib', alias='maximum message size',
        description='Maximum message size for a websocket message. '
        'Increase only if you get error messages or disconnects e.g. if you use large images.'
    )

    event_filter: WebsocketEventFilter = Field(
        default_factory=WebsocketEventFilter, alias='event filter',
        description='Configuration of server side event filters which will be applied to the websocket connection.'
    )

    ping_interval: int | float = Field(
        7, alias='ping interval', description='Interval for ping messages in seconds', gt=0
    )

    @field_validator('max_msg_size')
    def validate_see_buffer(cls, value: ByteSize):
        valid_values = (
            '128kib', '256kib', '512kib',
            '1Mib', '2Mib', '4Mib', '8Mib', '16Mib', '32Mib', '64Mib', '128Mib'
        )

        for _v in valid_values:
            # noinspection PyProtectedMember
            if value == ByteSize._validate(_v, None):
                return value

        msg = f'Value must be one of {", ".join(valid_values)}'
        raise ValueError(msg)


class Connection(BaseModel):
    url: str = Field(
        'http://localhost:8080', description='Connect to this url. Empty string ("") disables the connection.')
    user: str = ''
    password: str = ''
    verify_ssl: bool = Field(True, description='Check certificates when using https')

    websocket: Websocket = Field(
        default_factory=Websocket, in_file=False, description='Options for the websocket connection which is used'
        'to connect to the openHAB event bus.'
    )

    @field_validator('url')
    def validate_url(cls, value: str):
        if value:
            TypeAdapter(AnyHttpUrl).validate_python(value)
        return value


class OpenhabConfig(BaseModel):
    connection: Connection = Field(default_factory=Connection)
    general: General = Field(default_factory=General)
    ping: Ping = Field(default_factory=Ping)



from easyconfig import BaseModel
from pydantic import Field


class LocationConfig(BaseModel):
    """location where the instance is running. Is used to calculate Sunrise/Sunset and holidays."""

    latitude: float = Field(default=0.0)
    longitude: float = Field(default=0.0)
    elevation: float = Field(default=0.0)

    country: str = Field(default='', description='ISO 3166-1 Alpha-2 country code')
    subdivision: str = Field(
        default='', description='The subdivision (e.g. state or province) as a ISO 3166-2 code or its alias')



from __future__ import annotations

import logging
from typing import TypeAlias

from aiomqtt import Client, MqttError

import HABApp
from HABApp.core.connections import AutoReconnectPlugin, BaseConnection, Connections, ConnectionStateToEventBusPlugin
from HABApp.core.connections.base_connection import AlreadyHandledException
from HABApp.core.connections.base_plugin import BaseConnectionPluginConnectedTask


log = logging.getLogger('HABApp.mqtt.connection')


CONTEXT_TYPE: TypeAlias = Client | None


def setup() -> None:
    config = HABApp.config.CONFIG.mqtt

    from HABApp.mqtt.connection.handler import CONNECTION_HANDLER
    from HABApp.mqtt.connection.publish import PUBLISH_HANDLER
    from HABApp.mqtt.connection.subscribe import SUBSCRIPTION_HANDLER

    connection = Connections.add(CONNECTION)

    connection.register_plugin(CONNECTION_HANDLER, 0)
    connection.register_plugin(SUBSCRIPTION_HANDLER, 10)
    connection.register_plugin(PUBLISH_HANDLER, 20)

    connection.register_plugin(ConnectionStateToEventBusPlugin())
    connection.register_plugin(AutoReconnectPlugin())

    # config changes
    config.subscribe.subscribe_for_changes(SUBSCRIPTION_HANDLER.subscription_cfg_changed)
    config.connection.subscribe_for_changes(connection.status_configuration_changed)


class MqttConnection(BaseConnection):
    def __init__(self) -> None:
        super().__init__('mqtt')
        self.context: CONTEXT_TYPE = None

    def is_silent_exception(self, e: Exception):
        return isinstance(e, MqttError)


CONNECTION = MqttConnection()


class MqttPlugin(BaseConnectionPluginConnectedTask[MqttConnection]):

    def __init__(self, task_name: str) -> None:
        super().__init__(self._mqtt_wrap_task, task_name)

    async def mqtt_task(self):
        raise NotImplementedError()

    async def _mqtt_wrap_task(self) -> None:

        connection = self.plugin_connection
        log = connection.log
        log.debug(f'{self.task.name} task start')
        try:
            with connection.handle_exception(self.mqtt_task):
                await self.mqtt_task()
        except AlreadyHandledException:
            pass
        finally:
            log.debug(f'{self.task.name} task finished')



from .load_items import LoadOpenhabItemsPlugin
from .load_transformations import LoadTransformationsPlugin
from .out import (
    OUTGOING_PLUGIN,
    async_post_update,
    async_send_command,
    async_send_websocket_event,
    post_update,
    send_command,
    send_websocket_event,
)
from .overview_broken_links import BrokenLinksPlugin
from .overview_things import ThingOverviewPlugin
from .ping import PingPlugin
from .plugin_things import TextualThingConfigPlugin
from .wait_for_restore import WaitForPersistenceRestore
from .wait_for_startlevel import WaitForStartlevelPlugin
from .websockets import WebsocketPlugin



from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from immutables import Map
from typing_extensions import override

from HABApp.core.errors import InvalidItemValueError, ItemValueIsNoneError
from HABApp.openhab.definitions.websockets.item_value_types import (
    DecimalTypeModel,
    QuantityTypeModel,
)
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent


if TYPE_CHECKING:
    MetaData = MetaData     # noqa: PLW0127
    Mapping = Mapping       # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/NumberItem.java
class NumberItem(OpenhabItem):
    """NumberItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar int | float value: |oh_item_desc_value|
    :ivar str | None dimension: Dimension if it's a UoM item
    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('NumberItem', DecimalTypeModel, QuantityTypeModel, 'UnDef')
    _command_to_oh: Final = OutgoingCommandEvent('NumberItem', DecimalTypeModel, QuantityTypeModel, 'Refresh')

    def __init__(self, name: str, initial_value: int | float = None, label: str | None = None,
                 tags: frozenset[str] = frozenset(), groups: frozenset[str] = frozenset(),
                 metadata: Mapping[str, MetaData] = Map(), dimension: str | None = None) -> None:
        super().__init__(name, initial_value, label, tags, groups, metadata)
        self.dimension: str | None = dimension

    @override
    def _update_item_definition(self, item: 'NumberItem') -> None:
        super()._update_item_definition(item)
        self.dimension = item.dimension

    @staticmethod
    def _state_from_oh_str(state: str) -> int | float:
        if ' ' not in state:
            return DecimalTypeModel.get_value_from_state(state)
        return QuantityTypeModel.get_value_from_state(state)

    @property
    def unit(self) -> str | None:
        """Return the item unit if it is a "Unit of Measurement" item else None"""
        if (unit := self.metadata.get('unit')) is None:
            return None
        return unit.value

    def set_value(self, new_value: float | None) -> bool:

        if isinstance(new_value, (int, float)):
            return super().set_value(new_value)

        if new_value is None:
            return super().set_value(new_value)

        raise InvalidItemValueError.from_item(self, new_value)

    def __bool__(self) -> bool:
        if self.value is None:
            raise ItemValueIsNoneError.from_item(self)
        return bool(self.value)



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field, TypeAdapter


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/types/StateOption.java
class StateOptionResp(BaseModel):
    value: str
    label: str | None = None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/types/StateDescription.java
class StateDescriptionResp(BaseModel):
    minimum: int | float | None = None
    maximum: int | float | None = None
    step: int | float | None = None
    pattern: str | None = None
    read_only: bool = Field(alias='readOnly')
    options: tuple[StateOptionResp, ...]


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/types/CommandOption.java
class CommandOptionResp(BaseModel):
    command: str
    label: str | None = None


class CommandDescriptionResp(BaseModel):
    command_options: tuple[CommandOptionResp, ...] = Field(alias='commandOptions')


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/items/dto/GroupFunctionDTO.java
class GroupFunctionResp(BaseModel):
    name: str
    params: tuple[str, ...] = ()


class ItemResp(BaseModel):
    # ItemDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/items/dto/ItemDTO.java
    type: str
    name: str
    label: str | None = None
    category: str | None = None
    tags: tuple[str, ...]
    groups: tuple[str, ...] = Field(alias='groupNames')

    # EnrichedItemDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/item/EnrichedItemDTO.java
    link: str | None = None
    state: str
    transformed_state: str | None = Field(default=None, alias='transformedState')
    state_description: StateDescriptionResp | None = Field(default=None, alias='stateDescription')
    unit: str | None = Field(default=None, alias='unitSymbol')
    command_description: CommandDescriptionResp | None = Field(default=None, alias='commandDescription')
    metadata: dict[str, Any] = {}
    editable: bool = True

    # EnrichedGroupItemDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/item/EnrichedGroupItemDTO.java
    members: tuple[ItemResp, ...] = ()
    group_type: str | None = Field(default=None, alias='groupType')
    group_function: GroupFunctionResp | None = Field(default=None, alias='function')


ItemRespList = TypeAdapter(tuple[ItemResp, ...])


class ShortItemResp(BaseModel):
    type: str
    name: str
    state: str


ShortItemRespList = TypeAdapter(tuple[ShortItemResp, ...])



import logging
from collections.abc import Callable

# noinspection PyProtectedMember
from sys import _getframe as sys_get_frame
from types import FrameType, TracebackType
from typing import TYPE_CHECKING, Any, Final


if TYPE_CHECKING:
    import HABApp
    import HABApp.rule_manager

_NAME: Final = '__HABAPP__HOOK__'


log = logging.getLogger('HABApp.Rule')


class HABAppRuleHook:

    def __init__(self,
                 cb_register_rule: Callable[['HABApp.rule.Rule'], Any],
                 cb_suggest_name: Callable[['HABApp.rule.Rule'], str],
                 runtime: 'HABApp.runtime.Runtime', rule_file: 'HABApp.rule_manager.RuleFile') -> None:
        # callbacks
        self._cb_register: Final = cb_register_rule
        self._cb_suggest_name: Final = cb_suggest_name

        # runtime objs
        self.runtime: Final = runtime
        self.rule_file: Final = rule_file

        self.closed = False

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        self.closed = True

    def register_rule(self, rule: 'HABApp.rule.Rule'):
        if self.closed:
            # if we keep adding rules dynamically they will always get attached to the file and never unloaded
            log.warning(f'Added another rule of type {rule.__class__.__name__:s} '
                        'but file load has already been completed!')
        return self._cb_register(rule)

    def suggest_rule_name(self, rule: 'HABApp.rule.Rule') -> str:
        return self._cb_suggest_name(rule)

    def in_dict(self, obj: dict | None = None) -> dict:
        if obj is None:
            obj = {}
        obj[_NAME] = self
        return obj


def get_rule_hook() -> HABAppRuleHook:

    # noinspection PyUnresolvedReferences
    frame: FrameType | None = sys_get_frame(1)

    while frame is not None:
        _globals = frame.f_globals

        if (hook := _globals.get(_NAME)) is not None:
            assert isinstance(hook, HABAppRuleHook)
            return hook

        frame = frame.f_back

    msg = (
        'HABApp rule files are not meant to be executed directly! '
        'Put the file in the HABApp "rule" folder and HABApp will load it automatically.'
    )
    raise RuntimeError(msg)



from collections.abc import Mapping
from datetime import datetime
from typing import TYPE_CHECKING, Final

from HABApp.openhab.definitions.websockets.item_value_types import DateTimeTypeModel
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent


if TYPE_CHECKING:
    datetime = datetime     # noqa: PLW0127
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/DateTimeItem.java
class DatetimeItem(OpenhabItem):
    """DateTimeItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar datetime value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('DatetimeItem', DateTimeTypeModel, 'UnDef')
    _command_to_oh: Final = OutgoingCommandEvent('DatetimeItem', DateTimeTypeModel, 'Refresh')
    _state_from_oh_str: Final = staticmethod(DateTimeTypeModel.get_value_from_state)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from HABApp.openhab.definitions.websockets.item_value_types import RefreshTypeModel, StringTypeModel, UnDefTypeModel
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent


if TYPE_CHECKING:
    MetaData = MetaData     # noqa: PLW0127
    Mapping = Mapping       # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/StringItem.java
class StringItem(OpenhabItem):
    """StringItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar str value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    # Undef has to become before StringType
    _update_to_oh: Final = OutgoingStateEvent('StringItem', UnDefTypeModel, StringTypeModel)
    _command_to_oh: Final = OutgoingCommandEvent('StringItem', RefreshTypeModel, StringTypeModel)
    _state_from_oh_str: Final = staticmethod(StringTypeModel.get_value_from_state)



from pydantic import BaseModel, Field


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest/src/main/java/org/openhab/core/io/rest/internal/resources/beans/RootBean.java


class RuntimeResp(BaseModel):
    version: str
    build_string: str = Field(alias='buildString')


class LinkResp(BaseModel):
    type: str
    url: str


class RootResp(BaseModel):
    version: str
    locale: str
    measurement_system: str = Field(alias='measurementSystem')
    runtime_info: RuntimeResp = Field(alias='runtimeInfo')
    links: list[LinkResp]



from .event import EventFilter
from .groups import AndFilterGroup, OrFilterGroup
from .habapp_events import ValueChangeEventFilter, ValueCommandEventFilter, ValueUpdateEventFilter
from .no_filter import NoEventFilter



from collections.abc import Awaitable as __Awaitable
from collections.abc import Callable as __Callable
from typing import Any as __Any
from typing import Protocol as __Protocol
from typing import TypeAlias


TYPE_ANY_CLASS_TYPE: TypeAlias = type[object]
TYPE_FUNC_ASYNC: TypeAlias = __Callable[..., __Awaitable[__Any]]

TYPE_EVENT_CALLBACK: TypeAlias = __Callable[[__Any], __Any]


# noinspection PyPropertyDefinition
class HasNameAttr(__Protocol):
    @property
    def name(self) -> str:
        ...



from typing import Final

from fastnumbers import check_intlike
from fastnumbers import int as to_int
from immutables import Map
from javaproperties import loads as load_map_file

from HABApp.openhab.errors import MapTransformationNotFound
from HABApp.openhab.transformations._map.classes import (
    MapKeyType,
    MapTransformation,
    MapTransformationWithDefault,
    MapValueType,
)
from HABApp.openhab.transformations.base import TransformationFactoryBase, TransformationRegistryBase, log


class MapTransformationRegistry(TransformationRegistryBase):

    def __init__(self, name: str) -> None:
        super().__init__(name)
        self.objs: dict[str, tuple[Map[MapKeyType, MapValueType], MapValueType]] = {}

    def get(self, name: str) -> MapTransformation | MapTransformationWithDefault:
        try:
            data, default = self.objs[name]
        except KeyError:
            msg = f'Map transformation "{name:s}" not found!'
            raise MapTransformationNotFound(msg) from None

        if default is not None:
            return MapTransformationWithDefault(data, name=name, default=default)

        return MapTransformation(data, name=name)

    def set(self, name: str, configuration: dict[str, str]) -> None:

        if not (data := load_map_file(configuration['function'])):
            log.warning(f'Map transformation "{name:s}" is empty -> skipped!')
            return None

        default = data.pop('', None)
        if default is not None and check_intlike(default):
            default = to_int(default)

        map_keys = all(check_intlike(k) for k in data)
        map_values = all(check_intlike(v) for v in data.values())

        obj = {}
        for k, v in data.items():
            key = k if not map_keys else to_int(k)
            value = v if not map_values else to_int(v)
            obj[key] = value

        self.objs[name] = Map(obj), default


MAP_REGISTRY: Final = MapTransformationRegistry('map')


class MapTransformationFactory(TransformationFactoryBase[dict[MapKeyType, MapValueType]]):
    pass


MAP_FACTORY = MapTransformationFactory(MAP_REGISTRY)



import typing

from pydantic import BaseModel

from HABApp.core.const.topics import TOPIC_FILES
from HABApp.core.events.habapp_events import RequestFileLoadEvent
from HABApp.core.internals import uses_post_event


post_event = uses_post_event()

_PARAMETERS: dict[str, dict | list | BaseModel] = {}
_VALIDATORS: dict[str, BaseModel] = {}


def remove_parameter_file(file) -> None:
    _PARAMETERS.pop(file)


def set_parameter_file(file: str, value) -> None:
    # validate the parameters, this will raise an exception
    if model := _VALIDATORS.get(file):
        # validate and dump so we get the defaults
        value = model.model_validate(value).model_dump()

    _PARAMETERS[file] = value


def get_parameter_file(file: str):
    return _PARAMETERS[file]


def set_file_validator(filename: str, model: BaseModel | None) -> None:
    """Add a validator for the parameter file. If the file is already loaded this will reload the file.

    :param filename: filename which shall be validated (without extension)
    :param model   : Description of file content - see the library
                      `voluptuous <https://github.com/alecthomas/voluptuous#show-me-an-example/>`_ for examples.
                      Use `None` to remove validator.
    """

    # Remove validator
    if model is None:
        _VALIDATORS.pop(filename, None)
        return None

    if not isinstance(model, BaseModel):
        msg = f'Validator for {filename} must be an instance of BaseModel!'
        raise TypeError(msg)

    # Set validator
    old = _VALIDATORS.get(filename)
    _VALIDATORS[filename] = model

    if old is not None and old.model_json_schema() == model.model_json_schema():
        log.debug(f'Validator for {filename} did not change')
        return None

    log.debug(f'Validator for {filename} changed')
    post_event(TOPIC_FILES, RequestFileLoadEvent(filename))


def add_parameter(file: str, *keys, default_value):
    save = False

    if file not in _PARAMETERS:
        save = True
        param: typing.Dict[str, typing.Any] = {}
        _PARAMETERS[file] = param
    else:
        param = _PARAMETERS[file]

    if keys:
        # Create structure
        for key in keys[:-1]:
            if key not in param:
                param[key] = {}
                save = True
            param = param[key]

        # Create value
        if keys[-1] not in param:
            param[keys[-1]] = default_value
            save = True

    if save:
        save_file(file)
    return None


def get_value(file: str, *keys) -> typing.Any:
    try:
        param = _PARAMETERS[file]
    except KeyError:
        msg = f'File {file}.yml not found in params folder!'
        raise FileNotFoundError(msg) from None

    # lookup parameter
    for key in keys:
        param = param[key]
    return param


# Import here to prevent cyclic imports
from .parameter_files import log, save_file  # noqa: E402



from typing import Any, Literal

from pydantic import Field, Json
from typing_extensions import override

from HABApp.openhab.definitions import ThingStatusDetailEnum, ThingStatusEnum
from HABApp.openhab.events.thing_events import ThingAddedEvent as TargetThingAddedEvent
from HABApp.openhab.events.thing_events import ThingConfigStatusInfoEvent as TargetThingConfigStatusInfoEvent
from HABApp.openhab.events.thing_events import ThingFirmwareStatusInfoEvent as TargetThingFirmwareStatusInfoEvent
from HABApp.openhab.events.thing_events import ThingRemovedEvent as TargetThingRemovedEvent
from HABApp.openhab.events.thing_events import ThingStatusInfoChangedEvent as TargetThingStatusInfoChangedEvent
from HABApp.openhab.events.thing_events import ThingStatusInfoEvent as TargetThingStatusInfoEvent
from HABApp.openhab.events.thing_events import ThingUpdatedEvent as TargetThingUpdatedEvent

from .base import BaseEvent, BaseModel


class StatusPayload(BaseModel):
    status: ThingStatusEnum
    detail: ThingStatusDetailEnum = Field(alias='statusDetail')
    description: str = ''


class ThingStatusInfoEvent(BaseEvent):
    type: Literal['ThingStatusInfoEvent']
    payload: Json[StatusPayload]

    @override
    def to_event(self) -> TargetThingStatusInfoEvent:
        payload = self.payload
        return TargetThingStatusInfoEvent(
            name=self.topic[15:-7], status=payload.status, detail=payload.detail, description=payload.description
        )


class ThingStatusInfoChangedEvent(BaseEvent):
    type: Literal['ThingStatusInfoChangedEvent']
    payload: Json[tuple[StatusPayload, StatusPayload]]

    @override
    def to_event(self) -> TargetThingStatusInfoChangedEvent:
        new, old = self.payload
        return TargetThingStatusInfoChangedEvent(
            name=self.topic[15:-14],
            status=new.status, detail=new.detail, description=new.description,
            old_status=old.status, old_detail=old.detail, old_description=old.description
        )


class ConfigStatusPayloadMessageEntry(BaseModel):
    parameter: str = Field(alias='parameterName')
    type: str


class ConfigStatusPayload(BaseModel):
    messages: tuple[ConfigStatusPayloadMessageEntry, ...] = Field(alias='configStatusMessages')


class ThingConfigStatusInfoEvent(BaseEvent):
    type: Literal['ConfigStatusInfoEvent']
    payload: Json[ConfigStatusPayload]

    @override
    def to_event(self) -> TargetThingConfigStatusInfoEvent:
        messages = self.payload.messages
        return TargetThingConfigStatusInfoEvent(
            name=self.topic[15:-7],
            config_messages={m.parameter: m.type for m in messages}
        )


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/dto/ThingDTO.java#L27
class ThingRegistryPayload(BaseModel):
    name: str = Field(alias='UID')
    bridge_name: str | None = Field(None, alias='bridgeUID')
    type: str = Field(alias='thingTypeUID')
    label: str = Field(alias='label')
    location: str = Field('', alias='location')
    channels: tuple[dict[str, Any], ...] = ()
    configuration: dict[str, Any] = {}
    properties: dict[str, str] = {}


class ThingAddedEvent(BaseEvent):
    type: Literal['ThingAddedEvent']
    payload: Json[ThingRegistryPayload]

    @override
    def to_event(self) -> TargetThingAddedEvent:
        payload = self.payload
        return TargetThingAddedEvent(
            name=payload.name, thing_type=payload.type, label=payload.label, location=payload.location,
            channels=payload.channels, configuration=payload.configuration, properties=payload.properties
        )


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/events/ThingUpdatedEvent.java
class ThingUpdatedEvent(BaseEvent):
    type: Literal['ThingUpdatedEvent']
    payload: Json[tuple[ThingRegistryPayload, ThingRegistryPayload]]

    @override
    def to_event(self) -> TargetThingUpdatedEvent:
        payload, old = self.payload
        return TargetThingUpdatedEvent(
            name=payload.name, thing_type=payload.type, label=payload.label, location=payload.location,
            channels=payload.channels, configuration=payload.configuration, properties=payload.properties
        )


class ThingRemovedEvent(BaseEvent):
    type: Literal['ThingRemovedEvent']
    payload: Json[ThingRegistryPayload]

    @override
    def to_event(self) -> TargetThingRemovedEvent:
        payload = self.payload
        return TargetThingRemovedEvent(
            name=payload.name, thing_type=payload.type, label=payload.label, location=payload.location,
            channels=payload.channels, configuration=payload.configuration, properties=payload.properties
        )


class ThingFirmwareStatusInfoPayload(BaseModel):
    name: Any = Field(alias='thingUID')
    status: str = Field(alias='firmwareStatus')
    version: Any | None = None


class ThingFirmwareStatusInfoEvent(BaseEvent):
    type: Literal['FirmwareStatusInfoEvent']
    payload: Json[ThingFirmwareStatusInfoPayload]

    @override
    def to_event(self) -> TargetThingFirmwareStatusInfoEvent:
        payload = self.payload
        return TargetThingFirmwareStatusInfoEvent(
            name=self.topic[15:-16], status=payload.status
        )



from typing import Any, Final

from typing_extensions import Self

from HABApp.mqtt.interface_sync import publish


class MqttPublishOptions:
    """Allows to store the topic, qos and retain settings for a topic. These values can then be used to publish
    """
    def __init__(self, topic: str, qos: int | None = None, retain: bool | None = None) -> None:
        if not isinstance(topic, str):
            raise TypeError()
        if not topic:
            raise ValueError()

        self._topic: Final = topic
        self._qos: Final = qos
        self._retain: Final = retain

    @property
    def topic(self) -> str:
        """The topic"""
        return self._topic

    @property
    def qos(self) -> int | None:
        """QOS"""
        return self._qos

    @property
    def retain(self) -> bool | None:
        """Retain"""
        return self._retain

    def publish(self, payload: Any) -> None:
        """
        Publish a payload

        :param payload: MQTT Payload
        """

        return publish(self._topic, payload, qos=self._qos, retain=self._retain)

    def replace(self, topic: str | None = None, qos: int | None = None, retain: bool | None = None) -> Self:
        """
        Replace the topic, qos and retain with the given values and return a new object.

        :param topic: New topic (if provided)
        :param qos: New qos (if provided)
        :param retain: New retain (if provided)
        :return: New object
        """

        return self.__class__(
            topic if topic is not None else self._topic,
            qos if qos is not None else self._qos,
            retain if retain is not None else self._retain
        )



from . import item_value_types
from .all_events import (
    # Event Types
    OPENHAB_EVENT_TYPE,
    OPENHAB_EVENT_TYPE_ADAPTER,
)
from .item_events import ItemCommandSendEvent, ItemStateSendEvent
from .websocket_events import (
    WebsocketHeartbeatEvent,
    WebsocketRequestFailedEvent,
    WebsocketSendTypeFilter,
    WebsocketTopicEnum,
    WebsocketTypeFilterEvent,
)



from __future__ import annotations

from datetime import datetime
from typing import Any

from pydantic import BaseModel
from whenever import Instant, OffsetDateTime, PlainDateTime, SystemDateTime, ZonedDateTime

from HABApp.core.types import HSB, RGB
from HABApp.openhab.definitions.websockets.item_value_types import RawTypeModel


def convert_to_oh_str(obj: Any) -> str:
    if isinstance(obj, (str, int, float)):
        return str(obj)

    if isinstance(obj, datetime):
        # Add timezone (if not yet defined) to string, then remote anything below ms.
        # 2018-11-19T09:47:38.284000+0100 -> 2018-11-19T09:47:38.284+0100
        return obj.astimezone(None).strftime('%Y-%m-%dT%H:%M:%S.%f%z')

    if obj is None:
        return 'NULL'

    if isinstance(obj, RGB):
        obj = obj.to_hsb()

    if isinstance(obj, (list, tuple, set, frozenset)):
        return ','.join(convert_to_oh_str(x) for x in obj)

    if isinstance(obj, BaseModel):
        return obj.model_dump_json()

    if isinstance(obj, HSB):
        # noinspection PyProtectedMember
        return f'{obj._hue:.2f},{obj._saturation:.2f},{obj._brightness:.2f}'

    if isinstance(obj, bytes):
        return RawTypeModel.from_value(obj).value

    # https://whenever.readthedocs.io/en/latest/overview.html#iso-8601
    if isinstance(obj, (Instant, PlainDateTime, ZonedDateTime, OffsetDateTime, SystemDateTime)):
        return obj.format_common_iso()

    raise ValueError()


def map_null_str(value: str) -> str | None:
    if value in ('NULL', 'UNDEF'):
        return None
    return value



import asyncio
import contextlib
import logging
import re
from asyncio import Event, Task
from collections.abc import Awaitable, Callable
from pathlib import Path
from re import Pattern
from typing import Any, Final

from typing_extensions import override
from watchfiles import Change, DefaultFilter, awatch

from HABApp.core.asyncio import create_task_from_async
from HABApp.core.wrapper import process_exception


log = logging.getLogger('HABApp.file.events')
log.setLevel(logging.INFO)


DEFAULT_FILTER = DefaultFilter()

HABAPP_DISPATCHER_PREFIX: Final = 'HABAppInternal-'


class FileWatcherDispatcherBase:

    def __init__(self, name: str, coro: Callable[[str], Awaitable[Any]],) -> None:
        self._name: Final = name
        self._coro: Final = coro

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self._name:s}>'

    def __eq__(self, other: object) -> bool:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        return self._name

    def allow(self, change: Change | None, path: str) -> bool:
        raise NotImplementedError()

    async def dispatch(self, path: str) -> None:
        if not self.allow(None, path):
            return None

        try:
            await self._coro(path)
        except Exception as e:
            process_exception(self._coro, e, logger=log)


class FolderDispatcher(FileWatcherDispatcherBase):
    def __init__(self, name: str, coro: Callable[[str], Awaitable[Any]], folder: str) -> None:
        super().__init__(name, coro)
        self._folder: Final = folder

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FolderDispatcher):
            return False
        return self._name == other._name and self._coro is other._coro and self._folder == other._folder

    @override
    def allow(self, change: Change, path: str) -> bool:
        return path.startswith(self._folder)


class FileDispatcher(FileWatcherDispatcherBase):
    def __init__(self, name: str, coro: Callable[[str], Awaitable[Any]], file: str) -> None:
        super().__init__(name, coro)
        self._file: Final = file

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FileDispatcher):
            return False
        return self._name == other._name and self._coro is other._coro and self._file == other._file

    @override
    def allow(self, change: Change, path: str) -> bool:
        return path == self._file


class HABAppFileWatcher:
    def __init__(self) -> None:
        self._dispatchers: tuple[FileWatcherDispatcherBase, ...] = ()
        self._paths: tuple[str, ...] = ()
        self._files_task: Task | None = None
        self._stop_event: Final = Event()

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__:s}>'

    def cancel(self, dispatcher: FileWatcherDispatcherBase | str) -> None:
        if isinstance(dispatcher, str):
            for d in self._dispatchers:
                if d.name == dispatcher:
                    dispatcher = d
                    break
            else:
                msg = f'No dispatcher with name "{dispatcher:s}" found'
                raise ValueError(msg)

        self._dispatchers = tuple(d for d in self._dispatchers if d is not dispatcher)

    def watch_folder(self, name: str, coro: Callable[[str], Awaitable[Any]], folder: Path, *,
                     habapp_internal: bool = False) -> FolderDispatcher:
        return self._add_watcher(FolderDispatcher, name, coro, folder, habapp_internal=habapp_internal)

    def watch_file(self, name: str, coro: Callable[[str], Awaitable[Any]], file: Path, *,
                   habapp_internal: bool = False) -> FileDispatcher:
        return self._add_watcher(FileDispatcher, name, coro, file, habapp_internal=habapp_internal)

    def __notify_task(self) -> None:
        if self._files_task is None:
            self._files_task = create_task_from_async(self._watcher_task())
        else:
            self._stop_event.set()

    def _add_watcher(self, cls: type[FileDispatcher] | type[FolderDispatcher], name: str,
                     coro: Callable[[str], Awaitable[Any]], path: Path, *,
                     habapp_internal: bool = False) -> FileDispatcher | FolderDispatcher:
        d = cls(name if not habapp_internal else f'{HABAPP_DISPATCHER_PREFIX}{name}', coro, path.as_posix())
        self._add_dispatcher(d)
        self._add_path(path)
        return d

    def _add_dispatcher(self, dispatcher: FileWatcherDispatcherBase) -> None:
        name = dispatcher.name.lower()
        for d in self._dispatchers:
            if d.name.lower() != name or d == dispatcher:
                continue
            msg = f'Dispatcher with name "{dispatcher.name:s}" already exists'
            raise ValueError(msg)

        self._dispatchers += (dispatcher, )
        log.debug(f'Added dispatcher {dispatcher.name:s}')
        self.__notify_task()

    def _add_path(self, path: Path) -> None:
        if path.as_posix() in self._paths:
            return None

        if not path.is_dir() and not path.is_file():
            msg = f'Path {path} does not exist'
            raise FileNotFoundError(msg)

        self._paths += (path.as_posix(), )
        log.debug(f'Watching {path}')

        self.__notify_task()

    def _watch_filter(self, change: Change | None, path: str, *,
                      dispatchers: list[FileWatcherDispatcherBase] | None = None) -> bool:
        if not DEFAULT_FILTER(change, path):
            return False

        if dispatchers is not None:
            return any(dispatcher.allow(change, path) for dispatcher in dispatchers)

        process = any(dispatcher.allow(change, path) for dispatcher in self._dispatchers)
        log.debug(f'{change.name:s} {path:s}{" (ignored)" if not process else ""}')
        return process

    async def _watcher_task(self) -> None:
        delay = 1
        while self._dispatchers:
            await asyncio.sleep(1)

            try:
                self._stop_event.clear()
                log.debug('Starting file watcher')
                async for changes in awatch(*self._paths, debounce=120_000, step=1000,
                                            watch_filter=self._watch_filter, stop_event=self._stop_event):
                    file_names = sorted({Path(p).as_posix() for _, p in changes})
                    for dispatcher in self._dispatchers:
                        for path in file_names:
                            await dispatcher.dispatch(path)

                log.debug('File watcher stopped')
            except Exception as e:
                process_exception(self._watcher_task, e, logger=log)
                delay *= 2
                await asyncio.sleep(delay)

        log.debug('File watcher shutdown')

    async def shutdown(self) -> None:
        self._dispatchers = ()
        self._stop_event.set()
        if self._files_task is None:
            return None

        task = self._files_task
        self._files_task = None

        try:
            return await asyncio.wait_for(task, 2)
        except asyncio.TimeoutError:
            pass

        task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await task
        return None

    async def load_files(self, dispatcher_name_include: Pattern | str | None = None,
                         dispatcher_name_exclude: Pattern | str | None = None,
                         *,
                         exclude_habapp_config: bool = True) -> None:

        if isinstance(dispatcher_name_include, str):
            dispatcher_name_include = re.compile(dispatcher_name_include)
        if isinstance(dispatcher_name_exclude, str):
            dispatcher_name_exclude = re.compile(dispatcher_name_exclude)

        dispatchers = []
        for d in self._dispatchers:
            name = d.name

            if name.startswith(HABAPP_DISPATCHER_PREFIX):
                if exclude_habapp_config:
                    continue
                name = name.removeprefix(HABAPP_DISPATCHER_PREFIX)

            if exclude_habapp_config and d.name.lower().startswith(HABAPP_DISPATCHER_PREFIX):
                continue
            if dispatcher_name_include is not None and not dispatcher_name_include.search(name):
                continue
            if dispatcher_name_exclude is not None and dispatcher_name_exclude.search(name):
                continue
            dispatchers.append(d)

        if not dispatchers:
            msg = 'No dispatchers selected!'
            raise ValueError(msg)

        files: list[str] = []
        for path_str in self._paths:
            if not self._watch_filter(None, path_str, dispatchers=dispatchers):
                continue
            path = Path(path_str)

            if path.is_file():
                files.append(path.as_posix())
                continue

            if path.is_dir():
                for obj in path.glob('**/*'):
                    obj_str = obj.as_posix()
                    if self._watch_filter(None, obj_str, dispatchers=dispatchers):
                        files.append(obj_str)

        for file in sorted(files):
            for dispatcher in self._dispatchers:
                if dispatcher.allow(None, file):
                    await dispatcher.dispatch(file)



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Final

import HABApp
from HABApp.config import CONFIG
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.internals import uses_item_registry
from HABApp.openhab.connection.connection import OpenhabConnection
from HABApp.openhab.definitions.helpers.log_table import Table


if TYPE_CHECKING:
    from HABApp.openhab.definitions.rest import ThingResp


PING_CONFIG: Final = CONFIG.openhab.ping

Items = uses_item_registry()


class ThingOverviewPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self, name: str | None = None) -> None:
        super().__init__(name)
        self.do_run = True

    async def on_online(self) -> None:
        if not self.do_run:
            return None

        # don't run this after the connect, let the rules load etc.
        await asyncio.sleep(90)
        self.do_run = False

        self.draw_table(await HABApp.openhab.interface_async.async_get_things())

    @HABApp.core.wrapper.ignore_exception
    def draw_table(self, thing_data: tuple[ThingResp, ...]) -> None:

        thing_table = Table('Things')
        thing_stat = thing_table.add_column('Status', align='^')
        thing_label = thing_table.add_column('Label')
        thing_location = thing_table.add_column('Location')
        thing_type = thing_table.add_column('Thing type')
        thing_uid = thing_table.add_column('Thing UID')

        for node in thing_data:
            uid = node.uid
            type_uid = node.thing_type

            thing_uid.add(uid)
            thing_type.add(type_uid)
            thing_label.add(node.label)
            thing_stat.add(node.status.status)
            thing_location.add(node.location if node.location is not None else '')

        log = logging.getLogger('HABApp.openhab.things')
        for line in thing_table.get_lines(sort_columns=[thing_uid]):
            log.info(line)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Final

from HABApp.core.const import MISSING
from HABApp.core.errors import InvalidItemValueError, ItemValueIsNoneError
from HABApp.openhab.definitions.websockets.item_value_types import RawTypeModel
from HABApp.openhab.items.base_item import MetaData, OpenhabItem, OutgoingCommandEvent, OutgoingStateEvent
from HABApp.openhab.types import RawType


if TYPE_CHECKING:
    Mapping = Mapping       # noqa: PLW0127
    MetaData = MetaData     # noqa: PLW0127


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/library/items/ImageItem.java
class ImageItem(OpenhabItem):
    """ImageItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar RawType value: |oh_item_desc_value|

    :ivar str | None label: |oh_item_desc_label|
    :ivar frozenset[str] tags: |oh_item_desc_tags|
    :ivar frozenset[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    _update_to_oh: Final = OutgoingStateEvent('ImageItem', 'Raw', 'UnDef')
    _command_to_oh: Final = OutgoingCommandEvent('ImageItem', 'Refresh')
    _state_from_oh_str: Final = staticmethod(RawTypeModel.get_value_from_state)

    @property
    def image_type(self) -> str:
        """Image type (e.g. jpg or png)"""
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.type

    @property
    def image_bytes(self) -> bytes:
        """Image bytes"""
        if (v := self.value) is None:
            raise ItemValueIsNoneError.from_item(self)
        return v.data

    def set_value(self, new_value: RawType | tuple[str, bytes] | None) -> bool:

        if isinstance(new_value, RawType):
            return super().set_value(new_value)

        if isinstance(new_value, tuple):
            image_type, image_bytes = new_value
            if not image_type.startswith('image/') or not isinstance(image_bytes, bytes):
                raise InvalidItemValueError.from_item(self, new_value)
            return super().set_value(RawType.create(image_type, image_bytes))

        if new_value is None:
            return super().set_value(new_value)

        raise InvalidItemValueError.from_item(self, new_value)

    def oh_post_update(self, value: bytes | None = MISSING, image_type: str | None = None) -> None:
        """Post an update to an openHAB image with new image data. Image type is automatically detected,
        in rare cases when this does not work it can be set manually.

        :param value: image data
        :param image_type: (optional) what kind of image, ``jpeg`` or ``png``
        """
        if value is MISSING:
            value = self.value
        elif isinstance(value, bytes):
            value = RawType.create(data_type=image_type, data=value)
        return super().oh_post_update(value)

    def oh_send_command(self, value: bytes | None = MISSING, image_type: str | None = None) -> None:
        """Send a command to an openHAB image with new image data. Image type is automatically detected,
        in rare cases when this does not work it can be set manually.

        :param value: image data
        :param image_type: (optional) what kind of image, ``jpeg`` or ``png``
        """
        if value is MISSING:
            value = self.value
        elif isinstance(value, bytes):
            value = RawType.create(data_type=image_type, data=value)
        return super().oh_send_command(value)



import re
from pathlib import Path

from setuptools import find_packages, setup


THIS_FILE = Path(__file__)


# Load version number without importing HABApp
def load_version() -> str:
    version: dict[str, str] = {}
    exec((THIS_FILE.parent / 'src/HABApp/__version__.py').read_text(), version)  # noqa: S102
    assert version['__version__'], version
    return version['__version__']


def load_req() -> list[str]:

    lines = THIS_FILE.with_name('requirements_setup.txt').read_text().splitlines()
    print(lines)

    for i, line in enumerate(lines):
        if m := re.search(r'([^/]+).git@\w', line):
            lines[i] = f'{m.group(1):s} @ {line:s}'
            print(f'Modified: {lines[i]}')
    return lines


__version__ = load_version()

print(f'Version: {__version__}')
print()


# When we run tox tests we don't have these files available, so we skip them
readme = THIS_FILE.with_name('readme.md')
long_description = ''
if readme.is_file():
    with readme.open('r', encoding='utf-8') as fh:
        long_description = fh.read()

setup(
    name='HABApp',
    version=__version__,
    author='spaceman_spiff',
    # author_email="",
    description='Easy automation with MQTT and/or openHAB. Create home automation rules in python.',
    keywords=[
        'mqtt',
        'openhab',
        'habapp',
        'home automation'
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/spacemanspiff2007/HABApp',
    project_urls={
        'Documentation': 'https://habapp.readthedocs.io/',
        'GitHub': 'https://github.com/spacemanspiff2007/HABApp',
    },
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=['tests*']),
    package_data={'HABApp': ['py.typed']},
    install_requires=load_req(),
    python_requires='>=3.10',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: AsyncIO',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Home Automation'
    ],
    entry_points={
        'console_scripts': [
            'habapp = HABApp.__main__:main'
        ]
    }
)



from __future__ import annotations

from typing import Final

from typing_extensions import Self


class RawType:
    __slots__ = ('_type', '_data')  # noqa: RUF023

    def __init__(self, type: str, data: bytes) -> None:  # noqa: A002
        self._type: Final = type
        self._data: Final = data

    @property
    def type(self) -> str:
        return self._type

    @property
    def data(self) -> bytes:
        return self._data

    def __str__(self) -> str:
        data = self._data

        unit = 'kiB'
        size = len(self.data) / 1024
        if size > 1024:
            unit = 'MiB'
            size /= 1024
        fmt = '.1f' if size < 10 else '.0f'
        return (f'{self.__class__.__name__}(type={self._type:s} '
                f'data={data[:5].hex():s}..{data[-5:].hex():s} ({size:{fmt:s}}{unit}))')

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._type == other._type and self._data == other._data
        if isinstance(other, bytes):
            return self._data == other
        return NotImplemented

    @classmethod
    def create(cls, data_type: str | None, data: bytes) -> Self:
        if not isinstance(data, bytes):
            msg = f'Invalid data type! Expected bytes, got {type(data)}'
            raise TypeError(msg)

        if data_type is None:
            if data.startswith(b'\xFF\xD8\xFF'):
                data_type = 'jpeg'
            elif data.startswith(b'\x89\x50\x4E\x47'):
                data_type = 'png'
            elif data.startswith(b'\x47\x49\x46\x38'):
                data_type = 'gif'
            else:
                msg = f'Unknown image type! File Signature: {data[:10].hex()}'
                raise ValueError(msg)

        # Remove MIME type prefix for common image file types
        # https://developer.mozilla.org/en-US/docs/Web/Media/Formats/Image_types
        if data_type.endswith(('apng', 'avif', 'gif', 'jpeg', 'png', 'svg', 'webp', 'bmp', 'x-icon', 'tiff')):
            data_type = data_type.removeprefix('image/')

        return cls(data_type, data)



from datetime import timedelta

from easyconfig import BaseModel
from pydantic import Field, conint, model_validator
from typing_extensions import Self


class ThreadPoolConfig(BaseModel):
    enabled: bool = True
    '''When the thread pool is disabled HABApp will become an asyncio application.
    Use only if you have experience developing asyncio applications!
    If the thread pool is disabled using blocking calls in functions can and will break HABApp'''

    threads: conint(ge=1, le=32) = 10
    '''Amount of threads to use for the executor'''


class LoggingConfig(BaseModel):
    use_buffer: bool = Field(True, alias='use buffer')
    '''Automatically inject a buffer for the event log'''

    flush_every: float = Field(0.5, alias='flush every', ge=0.1)
    '''Wait time in seconds before the buffer gets flushed again when it was empty'''


class PeriodicTracebackDumpConfig(BaseModel):
    """Periodically dump the traceback of all currently running threads into a file"""

    enabled: bool = Field(False, description='Enable or disable functionality')

    delay: timedelta = Field('PT30M', gt=timedelta(seconds=0),
                             description='Initial delay before the first traceback dump')

    interval: timedelta = Field('PT1H', gt=timedelta(seconds=0),
                                description='Interval to dump the traceback')


class WatchEventLoopConfig(BaseModel):
    """Watch the asyncio event loop. If the loop is blocked dump the traceback of all running threads
    and shut down HABApp"""

    enabled: bool = Field(False, description='Enable or disable functionality')

    reset_every: timedelta = Field('PT1M', alias='reset every', gt=timedelta(seconds=0),
                                   description='Reset interval for the timeout')

    timeout: timedelta = Field('PT2M30S', gt=timedelta(seconds=0),
                               description='Timeout after which HABApp will shut down')

    @model_validator(mode='after')
    def _check_values(self) -> Self:
        # round to second
        self.timeout = timedelta(seconds=round(self.timeout.total_seconds()))
        self.reset_every = timedelta(seconds=round(self.reset_every.total_seconds()))

        if self.timeout <= self.reset_every:
            msg = f'Timeout must be greater than reset time! {self.timeout} > {self.reset_every}'
            raise ValueError(msg)
        return self


class DebugConfig(BaseModel):
    """Debugging options for HABApp"""

    periodic_traceback: PeriodicTracebackDumpConfig = Field(
        alias='periodic traceback', default_factory=PeriodicTracebackDumpConfig)

    traceback_on_shutdown_signal: bool = Field(
        False, alias='traceback on shutdown signal',
        description='Dump the traceback of all currently running threads into a file when receiving a shutdown signal. '
                    'Not available on Windows!'
    )

    watch_event_loop: WatchEventLoopConfig = Field(alias='watch event loop', default_factory=WatchEventLoopConfig)


class HABAppConfig(BaseModel):
    """HABApp internal configuration. Only change values if you know what you are doing!"""

    logging: LoggingConfig = Field(default_factory=LoggingConfig)
    thread_pool: ThreadPoolConfig = Field(default_factory=ThreadPoolConfig, alias='thread pool')
    debug: DebugConfig = Field(default_factory=DebugConfig)



from typing import Literal

from pydantic import Field, Json
from typing_extensions import override

from HABApp.openhab.events import ChannelDescriptionChangedEvent as TargetChannelDescriptionChangedEvent
from HABApp.openhab.events import ChannelTriggeredEvent as TargetChannelTriggeredEvent

from .base import BaseEvent, BaseModel


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/events/ChannelTriggeredEvent.java
class ChannelEventPayload(BaseModel):
    event: str
    channel: str


class ChannelTriggeredEvent(BaseEvent):
    type: Literal['ChannelTriggeredEvent']
    payload: Json[ChannelEventPayload]

    @override
    def to_event(self) -> TargetChannelTriggeredEvent:
        payload = self.payload
        return TargetChannelTriggeredEvent(
            name=self.topic[17:-10],
            event=payload.event,
            channel=payload.channel,
        )


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/events/ChannelDescriptionChangedEvent.java
class ChannelDescriptionChangedPayload(BaseModel):
    field: str
    name: str = Field(alias='channelUID')
    value: str
    old_value: str | None = Field(None, alias='oldValue')
    linked_items: tuple[str, ...] = Field((), alias='linkedItemNames')


class ChannelDescriptionChangedEvent(BaseEvent):
    type: Literal['ChannelDescriptionChangedEvent']
    payload: Json[ChannelDescriptionChangedPayload]

    @override
    def to_event(self) -> TargetChannelDescriptionChangedEvent:
        payload = self.payload
        return TargetChannelDescriptionChangedEvent(
            name=self.topic[17:-19],
            field=payload.field,
            value=payload.value,
        )




from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest/src/main/java/org/openhab/core/io/rest/internal/resources/beans/SystemInfoBean.java


class SystemInfoResp(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel)

    config_folder: str
    userdata_folder: str
    log_folder: str | None = None
    java_version: str | None = None
    java_vendor: str | None = None
    java_vendor_version: str | None = None
    os_name: str | None = None
    os_version: str | None = None
    os_architecture: str | None = None
    available_processors: int
    free_memory: int
    total_memory: int
    start_level: int

    uptime: int = -1    # TODO: remove default if we go OH4.1 only


class SystemInfoRootResp(BaseModel):
    system_info: SystemInfoResp = Field(alias='systemInfo')



from .base_event import OpenhabEvent
from .channel_events import ChannelDescriptionChangedEvent, ChannelTriggeredEvent
from .item_events import (
    GroupStateChangedEvent,
    GroupStateUpdatedEvent,
    ItemAddedEvent,
    ItemCommandEvent,
    ItemRemovedEvent,
    ItemStateChangedEvent,
    ItemStateEvent,
    ItemStatePredictedEvent,
    ItemStateUpdatedEvent,
    ItemUpdatedEvent,
)
from .thing_events import (
    ThingAddedEvent,
    ThingConfigStatusInfoEvent,
    ThingFirmwareStatusInfoEvent,
    ThingRemovedEvent,
    ThingStatusInfoChangedEvent,
    ThingStatusInfoEvent,
    ThingUpdatedEvent,
)


# isort: split

from .event_filters import (
    ItemCommandEventFilter,
    ItemStateChangedEventFilter,
    ItemStateEventFilter,
    ItemStateUpdatedEventFilter,
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from HABApp.core.internals import AutoContextBoundObj


if TYPE_CHECKING:
    from HABApp.util.multimode import MultiModeItem


class BaseMode(AutoContextBoundObj):

    def __init__(self, name: str) -> None:
        super().__init__()
        self.name: str = name

        self._mode_lower_prio: BaseMode | None = None
        self.parent: MultiModeItem | None = None

    def _set_mode_lower_prio(self, mode_lower_prio: BaseMode | None) -> None:
        assert isinstance(mode_lower_prio, BaseMode) or mode_lower_prio is None, type(mode_lower_prio)
        self._lower_priority_mode = mode_lower_prio

    def calculate_value(self, lower_prio_value: Any) -> Any:
        raise NotImplementedError()

    def cancel(self) -> None:
        """Remove the mode from the parent ``MultiModeItem`` and stop processing it
        """
        self._ctx_unlink()

        if (parent := self.parent) is not None:
            self.parent = None
            parent.remove_mode(self.name)

        self._mode_lower_prio = None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Final, Generic, ParamSpec, TypeVar

from HABApp.core.const.topics import TOPIC_ERRORS
from HABApp.core.events.habapp_events import HABAppException
from HABApp.core.internals import Context, ContextProvidingObj, uses_event_bus
from HABApp.core.lib import format_exception, get_obj_name


if TYPE_CHECKING:
    from collections.abc import Callable


default_logger = logging.getLogger('HABApp.Worker')

event_bus = uses_event_bus()


P = ParamSpec('P')
R = TypeVar('R')


class WrappedFunctionBase(ContextProvidingObj, Generic[P, R]):

    def __init__(self, func: Callable, name: str | None = None, logger: logging.Logger | None = None,
                 context: Context | None = None) -> None:

        # Allow setting of the rule context
        super().__init__(context)

        # name of the function
        if name is None:
            if self._habapp_ctx is not None:
                name = self._habapp_ctx.get_callback_name(func)
            if name is None:
                name = get_obj_name(func)
        self.name: Final[str] = name

        # Allow custom logger
        self.log: Final = default_logger if logger is None else logger

    def run(self, *args: P.args, **kwargs: P.kwargs) -> None:
        raise NotImplementedError()

    async def async_run(self, *args: P.args, **kwargs: P.kwargs) -> R | None:
        raise NotImplementedError()

    def process_exception(self, e: Exception, *args: Any, **kwargs: Any) -> None:

        lines = format_exception(e)

        # Log Exception
        self.log.error(f'Error in {self.name:s}: {e}')
        for line in lines:
            self.log.error(line)

        # Create an HABApp event, but only if we are not currently processing an exception while processing an error.
        # Otherwise, we might create an endless loop!
        if not args or not isinstance(args[0], HABAppException):
            event_bus.post_event(
                TOPIC_ERRORS, HABAppException(func_name=self.name, exception=e, traceback='\n'.join(lines))
            )
